#
#     This file is part of rockit.
#
#     rockit -- Rapid Optimal Control Kit
#     Copyright (C) 2019 MECO, KU Leuven. All rights reserved.
#
#     Rockit is free software; you can redistribute it and/or
#     modify it under the terms of the GNU Lesser General Public
#     License as published by the Free Software Foundation; either
#     version 3 of the License, or (at your option) any later version.
#
#     Rockit is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#     Lesser General Public License for more details.
#
#     You should have received a copy of the GNU Lesser General Public
#     License along with CasADi; if not, write to the Free Software
#     Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
#
#



from ...freetime import FreeTime
from ...casadi_helpers import prepare_build_dir, ConstraintInspector
from ..method import ExternalMethod, linear_coeffs, check_Js, reshape_number
from ...solution import OcpSolution

import casadi as ca
import numpy as np



def sx_write(expr):
    s = ca.symvar(expr)
    f = ca.Function('f',s,[expr])

    # Input values of the same dimensions as the above
    input_val = [e.name() for e in s]

    # Output values to be calculated of the same dimensions as the above
    output_val = [None]

    # Work vector
    work = [None]*f.sz_w()

    # For debugging
    instr = f.instructions_sx()

    # Loop over the algorithm
    for k in range(f.n_instructions()):

        # Get the atomic operation
        op = f.instruction_id(k)
        o = f.instruction_output(k)
        i = f.instruction_input(k)

        if(op==ca.OP_CONST):
            work[o[0]] = "%.16f" % float(f.instruction_constant(k))
        else:
            if op==ca.OP_INPUT:
                work[o[0]] = input_val[i[0]]
            elif op==ca.OP_OUTPUT:
                output_val[o[0]] = work[i[0]]
            elif op==ca.OP_SQ:
                work[o[0]] = work[i[0]] + "**2"
            else:
                disp_in = [work[a] for a in i]
                debug_str = ca.print_operator(instr[k],disp_in)
                work[o[0]] = debug_str
    return output_val[0]

def get_terms(e):
    def get_terms_internal(e):
        if e.op()==ca.OP_ADD:
            for i in range(e.n_dep()):
                for t in get_terms_internal(e.dep(i)):
                    yield t
        else:
            yield e
    return list(get_terms_internal(e))


def visit(e,parents=None):
    if parents is None:
        parents = []
    yield (e,parents)
    for i in range(e.n_dep()):
        for t in visit(e.dep(i),parents=[e]+parents):
            yield t

class CegarMethod(ExternalMethod):
    def __init__(self,
        method = None,
        **kwargs):
        self.method = method

        supported = {"free_T"}
        ExternalMethod.__init__(self, supported=supported, **kwargs)

    def to_sx(self,e,suffix=""):
        ret = []
        if e.is_scalar():
            ret.append(ca.SX.sym(e.name()+suffix))
        else:
            ret.extend(ca.SX.sym(e.name()+("_%d" % i) + suffix) for  i in range(e.numel()))
        return ret

    def to_sxs(self,L,suffix=""):
        ret = []
        for e in L:
            ret+=self.to_sx(e,suffix=suffix)
        return ret

    def get_state_names(self):
        ret = []
        for e in self.stage.states:
            if e.is_scalar():
                ret.append(e.name())
            else:
                ret.extend(e.name()+"_%d" % i for  i in range(e.numel()))
        return ret

    def get_control_names(self):
        ret = []
        for e in self.stage.controls:
            if e.is_scalar():
                ret.append(e.name())
            else:
                ret.extend(e.name()+"_%d" % i for  i in range(e.numel()))
        return ret

    def get_variable_names(self):
        return self.get_state_names()+self.get_control_names()

    def get_constants(self):
        ret = []
        for e in self.stage.parameters['']:
            if e.is_scalar():
                ret.append(e.name()+"=%.16f" % float(self.stage._param_vals[e]))
            else:
                nz = self.stage._param_vals[e].nonzeros()
                ret.extend(e.name()+"_%d=%.16f" % (i,nz[i]) for i in range(e.numel()))
        return ret

    def get_ODE(self):
        res = self.f(x=ca.vcat(self.x_args),u=ca.vcat(self.u_args),p=ca.vcat(self.p_args))["ode"]

        ret = []
        vars = self.get_state_names()
        for i,e in enumerate(vars):
            ret.append(e+"D' == " + sx_write(res[i]))

        return ret
    
    def get_cost(self):
        f = ca.Function('f',[ca.vcat(self.stage.states),ca.vcat(self.stage.controls),ca.vcat(self.stage.parameters[''])],[self.lagrange])
        result  = f(ca.vcat(self.x_args),ca.vcat(self.u_args),ca.vcat(self.p_args))
        return "costD' == " + sx_write(result)

    def get_path_constraints(self):

        # helpers to put limits on u
        ub_expr = []
        ub_l = []
        ub_u = []

        # Process path constraints
        for c, meta, args in self.stage._constraints["control"]:
            (lb,canon,ub),mc = self.constraint_inspector.canon(self.placeholders(c,preference=['expose']))
            # lb <= canon <= ub
            # Check for infinities
            try:
                lb_inf = np.all(np.array(ca.evalf(lb)==-np.inf))
            except:
                lb_inf = False
            try:
                ub_inf = np.all(np.array(ca.evalf(ub)==np.inf))
            except:
                ub_inf = False

            if mc.type == ca.OPTI_EQUALITY:
                assert "Not supported"
            else:
                assert mc.type in [ca.OPTI_INEQUALITY, ca.OPTI_GENERIC_INEQUALITY, ca.OPTI_DOUBLE_INEQUALITY]

                # Catch simple bounds on u
                if ca.is_linear(canon, self.stage.u) and not ca.depends_on(canon, ca.vertcat(self.stage.x, self.v)):
                    J,c = ca.linear_coeff(canon, self.stage.u)
                    try:
                        check_Js(J)
                        ub_expr.append(J)
                        if ub_inf:
                            ub_u.append(reshape_number(J @ self.stage.u,-np.inf))
                        else:
                            ub_u.append(ub-c)
                        if lb_inf:
                            ub_l.append(reshape_number(J @ self.stage.u,np.inf))
                        else:
                            ub_l.append(lb-c)
                        continue
                    except:
                        pass

        ub_expr = ca.evalf(ca.vcat(ub_expr))
        ub_l = ca.evalf(ca.vcat(ub_l))
        ub_u = ca.evalf(ca.vcat(ub_u))
        # Add missing rows
        rows = set(ca.sum1(ub_expr).T.row())
        missing_rows = [i for i in range(self.stage.nu) if i not in rows]
        M = ca.DM(len(missing_rows), self.stage.nu)
        for i,e in enumerate(missing_rows):
            M[i,e] = 1

        ub_expr = ca.vertcat(ub_expr,M)
        ub_l = ca.vertcat(ub_l,-np.inf*ca.DM.ones(len(missing_rows)))
        ub_u = ca.vertcat(ub_u,np.inf*ca.DM.ones(len(missing_rows)))
       
        ub_l = ca.solve(ub_expr,ub_l)
        ub_u = ca.solve(ub_expr,ub_u)

        f = ca.Function('f',[ca.vcat(self.stage.parameters[''])],[ub_l,ub_u])
        result  = f(ca.vcat(self.p_args))
        ret = []
        for i, n in enumerate(self.get_control_names()):
            lb = sx_write(result[0][i])
            ub = sx_write(result[1][i])
            ret.append("%s <= (%s <= %s)" % (lb, n, ub))
        return ret
    
    def get_initial_constraints(self):

        x0_eq = []
        x0_b = []

        # Process point constraints
        # Probably should de-duplicate stuff wrt path constraints code
        for c, meta, _ in self.stage._constraints["point"]:
            # Make sure you resolve u to r_at_t0/r_at_tf
            c = self.placeholders(c,max_phase=1)
            has_t0 = 'r_at_t0' in [a.name() for a in ca.symvar(c)]
            has_tf = 'r_at_tf' in [a.name() for a in ca.symvar(c)]

            cb = c
            (lb,canon,ub),mc = self.constraint_inspector.canon(self.placeholders(c,preference=['expose']))

            if has_t0:
                # t0
                check = ca.is_linear(canon, self.stage.x)
                check = check and not ca.depends_on(canon, ca.vertcat(self.stage.u, self.v))
                assert check and mc.type == ca.OPTI_EQUALITY, "at t=t0, only equality constraints on x are allowed. Got '%s'" % str(c)

                J,c = ca.linear_coeff(canon, self.stage.x)
                try:
                    J = ca.evalf(J)
                    x0_eq.append(J)
                    x0_b.append(lb-c)
                    continue
                except:
                    pass

        x0_eq = ca.vcat(x0_eq)
        x0_b = ca.vcat(x0_b)
        x_current = ca.inv(ca.evalf(x0_eq)) @ x0_b
        f = ca.Function('f',[ca.vcat(self.stage.parameters[''])],[x_current])
        result  = f(ca.vcat(self.p_args))
        ret = []

        for i, n in enumerate(self.get_state_names()):
            ret.append(n+" = " + sx_write(result[i]))
            
        return ret

    def get_sample_time(self):
        #dt = self.method.control_grid[1]-self.method.control_grid[0]
        dt = self.stage._T/self.method.N#control_grid[1]-self.method.control_grid[0]
        return "sample_time = %.16f" % float(dt)

    def fill_placeholders_integral(self, phase, stage, expr, *args):
        if phase==1:
            return expr

    def fill_placeholders_sum_control(self, phase, stage, expr, *args):
        raise Exception("ocp.sum not supported. Use ocp.integral instead.")

    def fill_placeholders_sum_control_plus(self, phase, stage, expr, *args):
        raise Exception("ocp.sum not supported. Use ocp.integral instead.")

    def transcribe_phase1(self, stage, **kwargs):
        self.stage = stage

        self.v = ca.vvcat(stage.variables[''])
        self.X_gist = [ca.MX.sym("Xg", stage.nx) for k in range(self.N+1)]
        self.U_gist = [ca.MX.sym("Ug", stage.nu) for k in range(self.N)]
        self.V_gist = ca.MX.sym("Vg", *self.v.shape)
        self.T_gist = ca.MX.sym("Tg")

        self.constraint_inspector = ConstraintInspector(self, stage)


        #self.time_grid = self.grid(stage._t0, stage._T, self.N)
        self.normalized_time_grid = self.grid(0.0, 1.0, self.N)
        self.time_grid = self.normalized_time_grid
        if not isinstance(stage._T, FreeTime): self.time_grid*= stage._T
        if not isinstance(stage._t0, FreeTime): self.time_grid+= stage._t0
        self.control_grid = ca.MX(stage.t0 + self.normalized_time_grid*stage.T).T

        self.lagrange = ca.MX(0)
        self.mayer = ca.MX(0)
        var_mayer = []
        obj = ca.MX(stage._objective)
        terms = get_terms(obj)
        for term in terms:
            n = [e.name() for e in ca.symvar(term)]
            sumi = np.sum([e=="r_integral" for e in n])
            summ = np.sum([e=="r_at_tf" for e in n])
            if sumi>1:
                raise Exception("Objective cannot be parsed: operation combining two integrals not supported")
            if sumi==1:
                n_hits = 0
                for e,parents in visit(term):
                    if e.is_symbolic() and e.name()=="r_integral":
                        n_hits+=1
                        for pi in range(len(parents)):
                            p = parents[pi]
                            pchild = parents[pi-1] if pi>0 else e
                            correct = False
                            # Distributive operations
                            if p.op()==ca.OP_MUL:
                                correct = True
                            if p.op()==ca.OP_DIV:
                                # Only allow integral in LHS
                                correct = hash(p.dep(0))==hash(pchild)
                            assert correct, "Objective cannot be parsed: integrals can only be multiplied or divided."
                    if n_hits>1:
                        raise Exception("Objective cannot be parsed")
                if summ!=0:
                    raise Exception("Objective cannot be parsed: operation combining integral and at_tf part not supported")
                self.lagrange += term
                continue
            self.mayer += term

    def transcribe_phase2(self, stage, **kwargs):

        self.constraint_inspector.finalize()
        
        placeholders = kwargs["placeholders"]



        # Total Lagrange integrand
        self.lagrange = placeholders(self.lagrange,preference=['expose'])
        # Total Mayer term
        self.mayer = placeholders(self.mayer,preference=['expose'])

        self.f = self.stage._ode()
        self.x_args = self.to_sxs(self.stage.states,suffix="D")
        self.u_args = self.to_sxs(self.stage.controls,suffix="D")
        self.p_args = self.to_sxs(self.stage.parameters[''])
        self.placeholders = placeholders




    def to_function(self, stage, name, args, results, *margs):
        print("args=",args)

        res = self.solver(p=stage.p)
        print(stage.p)
        print([stage.value(a) for a in args])


        [_,states] = stage.sample(stage.x,grid='control')
        [_,controls] = stage.sample(stage.u,grid='control-')
        variables = stage.value(vvcat(stage.variables['']))

        helper_in = [states,controls,variables, stage.T]
        helper = Function("helper", helper_in, results)

        arg_in = helper(res["x_opt"],res["u_opt"],res["v_opt"],res["T_opt"])

        ret = Function(name, args, arg_in, *margs)
        assert not ret.has_free()
        return ret

    def initial_value(self, stage, expr):
        ret = self.pmap(p=self.P0)
        parameters = []
        for p in stage.parameters['']:
            parameters.append(stage.value(p))
        
        [_,states] = stage.sample(stage.x,grid='control')
        [_,controls] = stage.sample(stage.u,grid='control-')
        variables = stage.value(vvcat(stage.variables['']))

        helper_in = [vvcat(parameters),states,controls,variables, stage.T]
        helper = Function("helper", helper_in, [expr])
        return helper(self.P0, cs.repmat(ret["x_current"], 1, self.N+1), cs.repmat(self.U0, 1, self.N), self.V0, 0).toarray(simplify=True)

    def solve(self, stage,limited=False):
        self.solver.generate_in(self.build_dir_abs+os.sep+"debug_in.txt",[self.P0])
        ret = self.solver(p=self.P0)


        self.stats = stats = self._grampc_driver_get_stats_internal().contents
        self.last_solution = OcpSolution(SolWrapper(self, vec(ret["x_opt"]), vec(ret["u_opt"]), ret["v_opt"], ret["T_opt"], rT=stage.T), stage)

        conv = stats.conv_grad and stats.conv_con

        if not conv:
            if stats.n_outer_iter==self.grampc_options["MaxMultIter"]:
                if not limited:
                    raise Exception("MaxMultIter exhausted without meeting convergence criteria")
            else:
                raise Exception("Problem not converged")

        return self.last_solution

    def get_stats(self):
        stats = self.stats
        return dict((k,getattr(stats,k)) for k,_ in stats_fields)

    def non_converged_solution(self, stage):
        return self.last_solution

    def solve_limited(self, stage):
        return self.solve(stage,limited=True)

    def eval(self, stage, expr):
        placeholders = stage.placeholders_transcribed
        expr = placeholders(expr,max_phase=1)
        ks = [self.v,stage.T]
        vs = [self.V_gist, self.T_gist]
        ret = substitute([expr],ks,vs)[0]
        return ret
        
    @property
    def gist(self):
        return vertcat(ExternalMethod.gist.fget(self), self.V_gist, self.T_gist)

    def eval_at_control(self, stage, expr, k):
        placeholders = stage.placeholders_transcribed
        expr = placeholders(expr,max_phase=1)
        ks = [stage.x,stage.u,self.v,stage.T]
        vs = [self.X_gist[k], self.U_gist[min(k, self.N-1)], self.V_gist, self.T_gist]
        if not self.t_state:
            ks += [stage.t]
            vs += [self.control_grid[k]]
        ret = substitute([expr],ks,vs)[0]
        return ret

class SolWrapper:
    def __init__(self, method, x, u, v, T, rT=None):
        self.method = method
        self.x = x
        self.u = u
        self.T = T
        self.v = v
        self.rT = rT

    def value(self, expr, *args,**kwargs):
        placeholders = self.method.stage.placeholders_transcribed
        expr = substitute(expr,self.rT, self.T)
        ret = evalf(substitute([placeholders(expr)],[self.method.gist],[vertcat(self.x, self.u, self.v, self.T)])[0])
        return ret.toarray(simplify=True)

    def stats(self):
        return self.method.get_stats()



# -*- coding: future_fstrings -*-
#
# Copyright (c) The acados authors.
#
# This file is part of acados.
#
# The 2-Clause BSD License
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
# this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.;
#

from copy import deepcopy
from typing import Tuple
import casadi as ca
import numpy as np

from .acados_model import AcadosModel
from .acados_ocp import AcadosOcp, AcadosOcpConstraints
from .utils import casadi_length, is_empty


def create_model_with_cost_state(ocp: AcadosOcp) -> Tuple[AcadosModel, np.ndarray]:
    """
    Creates a new AcadosModel with an extra state `cost_state`,
    which has the dynamics of the cost function and slack penalties corresponding to the intermediate shooting nodes.

    Note: In contrast cost_discretization='INTEGRATOR', this allows to integrate also the slack penalties.
    Since l1 slack penalties are nondifferentiable, an accurate cost integration with the model created by this function should use many integrator steps, when slack penalties are part of the OCP formulation.

    Returns the augmented model and the parameter values of the given AcadosOcp.
    """

    model = deepcopy(ocp.model)
    symbol = model.get_casadi_symbol()
    cost_state = symbol("cost_state")
    cost_state_dot = symbol("cost_state_dot")

    if ocp.cost.cost_type == "LINEAR_LS":
        y = ocp.cost.Vx @ model.x + ocp.cost.Vu @ model.u
        if casadi_length(model.z) > 0:
            ocp.cost.Vz @ model.z
        residual = y - ocp.cost.yref
        cost_dot = 0.5*(residual.T @ ocp.cost.W @ residual)

    elif ocp.cost.cost_type == "NONLINEAR_LS":
        residual = model.cost_y_expr - ocp.cost.yref
        cost_dot = 0.5*(residual.T @ ocp.cost.W @ residual)

    elif ocp.cost.cost_type == "EXTERNAL":
        cost_dot = model.cost_expr_ext_cost

    elif ocp.cost.cost_type == "CONVEX_OVER_NONLINEAR":
        cost_dot = ca.substitute(
            model.cost_psi_expr, model.cost_r_in_psi_expr, model.cost_y_expr)

    else:
        raise Exception("create_model_with_cost_state: Unknown cost type.")

    i_slack = 0
    for ibu in ocp.constraints.idxsbu:
        iu = ocp.constraints.idxbu[ibu]
        lower_violation = ca.fmax(ocp.constraints.lbu[ibu] - model.u[iu], 0)
        upper_violation = ca.fmax(model.u[iu] - ocp.constraints.ubu[ibu], 0)
        cost_dot += ocp.cost.zl[i_slack] * lower_violation + \
            ocp.cost.Zl[i_slack] * lower_violation ** 2
        cost_dot += ocp.cost.zu[i_slack] * upper_violation + \
            ocp.cost.Zu[i_slack] * upper_violation ** 2
        i_slack += 1

    for ibx in ocp.constraints.idxsbx:
        ix = ocp.constraints.idxbx[ibx]
        lower_violation = ca.fmax(ocp.constraints.lbx[ibx] - model.x[ix], 0)
        upper_violation = ca.fmax(model.x[ix] - ocp.constraints.ubx[ibx], 0)
        cost_dot += ocp.cost.zl[i_slack] * lower_violation + \
            ocp.cost.Zl[i_slack] * lower_violation ** 2
        cost_dot += ocp.cost.zu[i_slack] * upper_violation + \
            ocp.cost.Zu[i_slack] * upper_violation ** 2
        i_slack += 1

    if not is_empty(ocp.constraints.C):
        g = ocp.constraints.C @ ocp.model.x + ocp.constraints.D @ ocp.model.u
        for ig in ocp.constraints.idxsg:
            lower_violation = ca.fmax(ocp.constraints.lg[ig] - g[ig], 0)
            upper_violation = ca.fmax(g[ig] - ocp.constraints.ug[ig], 0)
            cost_dot += ocp.cost.zl[i_slack] * lower_violation + \
                ocp.cost.Zl[i_slack] * lower_violation ** 2
            cost_dot += ocp.cost.zu[i_slack] * upper_violation + \
                ocp.cost.Zu[i_slack] * upper_violation ** 2
            i_slack += 1

    for ih in ocp.constraints.idxsh:
        lower_violation = ca.fmax(
            ocp.constraints.lh[ih] - ocp.model.con_h_expr[ih], 0)
        upper_violation = ca.fmax(
            ocp.model.con_h_expr[ih] - ocp.constraints.uh[ih], 0)
        cost_dot += ocp.cost.zl[i_slack] * lower_violation + \
            ocp.cost.Zl[i_slack] * lower_violation ** 2
        cost_dot += ocp.cost.zu[i_slack] * upper_violation + \
            ocp.cost.Zu[i_slack] * upper_violation ** 2
        i_slack += 1

    if not is_empty(ocp.constraints.idxsphi):
        raise NotImplementedError(
            f"Not implemented for nontrivial ocp.constraints.idxsphi = {ocp.constraints.idxsphi}")

    model.x = ca.vertcat(model.x, cost_state)
    model.xdot = ca.vertcat(model.xdot, cost_state_dot)
    model.f_expl_expr = ca.vertcat(model.f_expl_expr, cost_dot)
    model.f_impl_expr = ca.vertcat(model.f_impl_expr, cost_state_dot-cost_dot)

    return model, ocp.parameter_values


def detect_constraint_structure(model: AcadosModel, constraints: AcadosOcpConstraints, stage_type: str):
    """
    - stage_type: allowed values "initial", "path", "terminal"
    """
    x = model.x
    u = model.u
    z = model.z
    nx = x.shape[0]
    nu = u.shape[0]

    p_global = model.p_global

    casadi_var = model.get_casadi_symbol()
    casadi_dm_zeros = ca.DM.zeros

    if stage_type == 'initial':
        expr_constr = model.con_h_expr_0
        lb = constraints.lh_0
        ub = constraints.uh_0
        print('\nConstraint detection for initial constraints.')
    elif stage_type == 'path':
        expr_constr = model.con_h_expr
        lb = constraints.lh
        ub = constraints.uh
        print('\nConstraint detection for path constraints.')
    elif stage_type == 'terminal':
        expr_constr = model.con_h_expr_e
        lb = constraints.lh_e
        ub = constraints.uh_e
        print('\nConstraint detection for terminal constraints.')
    else:
        raise ValueError('Constraint detection: Wrong stage_type.')

    if is_empty(expr_constr):
        expr_constr = casadi_var('con_h_expr', 0, 0)

    # Initialize
    constr_expr_h = casadi_var('con_h_expr', 0, 0)
    lh = []
    uh = []

    c_lin = casadi_dm_zeros(0, nx)
    d_lin = casadi_dm_zeros(0, nu)
    lg = []
    ug = []

    Jbx = casadi_dm_zeros(0, nx)
    lbx = []
    ubx = []

    Jbu = casadi_dm_zeros(0, nu)
    lbu = []
    ubu = []

    # Loop over CasADi formulated constraints
    for ii in range(expr_constr.shape[0]):
        c = expr_constr[ii]
        if any(ca.which_depends(c, z)) or \
                not ca.is_linear(c, ca.vertcat(x, u)) or \
                any(ca.which_depends(c, model.p)) or \
                any(ca.which_depends(c, p_global)):

            # External constraint
            constr_expr_h = ca.vertcat(constr_expr_h, c)
            lh.append(lb[ii])
            uh.append(ub[ii])
            print(f'constraint {ii+1} is kept as nonlinear constraint.')
            print(c)
            print(' ')
        else:  # c is linear in x and u
            Jc_fun = ca.Function('Jc_fun', [x[0]], [
                                 ca.jacobian(c, ca.vertcat(x, u))])
            Jc = Jc_fun(0)

            if np.sum(Jc != 0) == 1:
                # c is bound
                idb = Jc.full().squeeze().nonzero()[0][0]
                if idb < nx:
                    # Bound on x
                    Jbx = ca.vertcat(Jbx, casadi_dm_zeros(1, nx))
                    Jbx[-1, idb] = 1
                    lbx.append(lb[ii] / Jc[idb])
                    ubx.append(ub[ii] / Jc[idb])
                    print(f'constraint {ii+1} is reformulated as bound on x.')
                    print(c)
                    print(' ')
                else:
                    # Bound on u
                    Jbu = ca.vertcat(Jbu, casadi_dm_zeros(1, nu))
                    Jbu[-1, idb - nx] = 1
                    lbu.append(lb[ii] / Jc[idb])
                    ubu.append(ub[ii] / Jc[idb])
                    print(f'constraint {ii+1} is reformulated as bound on u.')
                    print(c)
                    print(' ')
            else:
                # c is general linear constraint
                c_lin = ca.vertcat(c_lin, Jc[:nx])
                d_lin = ca.vertcat(d_lin, Jc[nx:])
                lg.append(lb[ii])
                ug.append(ub[ii])
                print(
                    f'constraint {ii+1} is reformulated as general linear constraint.')
                print(c)
                print(' ')

    if stage_type == 'terminal':
        # Checks
        if any(ca.which_depends(expr_constr, u)) or lbu or (d_lin.size()[0] > 0 and any(d_lin)):
            raise ValueError(
                'Terminal constraint may not depend on control input.')
        # h
        constraints.constr_type_e = 'BGH'
        if lh:
            model.con_h_expr_e = constr_expr_h
            constraints.lh_e = np.array(lh)
            constraints.uh_e = np.array(uh)
        else:
            model.con_h_expr_e = None
            constraints.lh_e = np.array([])
            constraints.uh_e = np.array([])
        # linear constraint g
        if lg:
            constraints.C_e = c_lin
            constraints.lg_e = np.array(lg)
            constraints.ug_e = np.array(ug)
        # Bounds x
        if lbx:
            constraints.idxbx_e = J_to_idx(Jbx)
            constraints.lbx_e = np.array(lbx)
            constraints.ubx_e = np.array(ubx)

    elif stage_type == 'initial':
        print("At initial stage, only h constraints are detected.")
        constraints.constr_type_0 = 'BGH'
        # h
        if lh:
            model.con_h_expr_0 = constr_expr_h
            constraints.lh_0 = np.array(lh)
            constraints.uh_0 = np.array(uh)
        else:
            model.con_h_expr_0 = None
            constraints.lh_0 = np.array([])
            constraints.uh_0 = np.array([])
    else:  # path
        constraints.constr_type = 'BGH'
        # nonlinear constraint
        if lh:
            model.con_h_expr = constr_expr_h
            constraints.lh = np.array(lh)
            constraints.uh = np.array(uh)
        else:
            model.con_h_expr = None
            constraints.lh = np.array([])
            constraints.uh = np.array([])
        # linear constraint g
        if lg:
            constraints.C = np.array(c_lin)
            constraints.D = np.array(d_lin)
            constraints.lg = np.array(lg)
            constraints.ug = np.array(ug)
        # Bounds x
        if lbx:
            constraints.idxbx = J_to_idx(Jbx)
            constraints.lbx = np.array(lbx)
            constraints.ubx = np.array(ubx)
        # Bounds u
        if lbu:
            constraints.idxbu = J_to_idx(Jbu)
            constraints.lbu = np.array(lbu)
            constraints.ubu = np.array(ubu)


def J_to_idx(J):
    nrows = J.size()[0]
    idx = []
    for i in range(nrows):
        this_idx = ca.DM(J[i, :].sparsity()).full().nonzero()[0]
        if len(this_idx) != 1:
            raise ValueError(
                f'J_to_idx: Invalid J matrix. Exiting. Found more than one nonzero in row {i+1}.')
        if J[i, this_idx] != 1:
            raise ValueError(
                f'J_to_idx: J matrices can only contain 1s, got J({i+1}, {this_idx}) = {J[i, this_idx]}')
        idx.append(this_idx[0])
    return np.array(idx)



import unittest

from rockit import Ocp, DirectMethod, MultipleShooting, SingleShooting, FreeTime, Stage, GeometricGrid
import numpy as np
from casadi import kron, DM, vertcat

from problems import bang_bang_problem
from rockit.spline_method import SplineMethod

class StageTests(unittest.TestCase):

    def test_stage_next(self):
      for method in [MultipleShooting(N=10),SplineMethod(N=10)]:
        (ocp, p, v, u) = bang_bang_problem(method)

        ocp.subject_to(-0.3 <= (ocp.next(u)-u <=0.3)  )
        #ocp.subject_to(-0.3 <= ((ocp.next(u)-u)/(ocp.next(ocp.t) - ocp.t)<=0.3) )

        sol = ocp.solve()

        usol = sol.sample(u,grid='control')[1]

        self.assertAlmostEqual(np.linalg.norm(np.diff(usol,axis=0),np.inf), 0.3, places=5)

    def test_inf_der(self):
      (ocp, p, v, u) = bang_bang_problem(MultipleShooting(N=10))

      ocp.subject_to(-0.3 <= (ocp.inf_der(p) <= 0.3),grid='inf' )

      sol = ocp.solve()

      vsol = sol.sample(v,grid='control')[1]
      self.assertAlmostEqual(np.linalg.norm(vsol,np.inf),0.3, places=5)

    def test_objective_next(self):
      ref = None
      for method in [MultipleShooting(N=10),SplineMethod(N=10)]:
        (ocp, p, v, u) = bang_bang_problem(method)

        #p_ref = ocp.parameter(2, grid='control')
        #ocp.set_value(p_ref, 1)
        
        x = vertcat(p,v)
        dx = ocp.next(x)-x
        ocp.add_objective(ocp.sum(dx.T @ dx))

        sol = ocp.solve()
        if ref is None:
            ref = sol.value(ocp.T)
        else:
            np.testing.assert_allclose(ref,sol.value(ocp.T))


    def test_set_next(self):
      ocp = Ocp(T=10)

      x1 = ocp.state()
      x2 = ocp.state()

      u = ocp.control()

      ocp.set_next(x1, (1 - x2**2) * x1 - x2 + u)
      ocp.set_next(x2, x1)

      ocp.add_objective(ocp.sum(x1**2 + x2**2 + u**2))

      ocp.subject_to(-1 <= (u <= 1))
      ocp.subject_to(x1 >= -0.25)

      ocp.subject_to(ocp.at_t0(x1) == 0)
      ocp.subject_to(ocp.at_t0(x2) == 1)

      ocp.solver('ipopt')

      N = 10
      for M in [1,2]:
        ocp.method(MultipleShooting(N=N,M=M))
        
        sol = ocp.solve()

        x1sol = sol.sample(x1, grid='integrator')[1]
        x2sol = sol.sample(x2, grid='integrator')[1]
        usol = kron(DM.ones(1,M),sol.sample(u, grid='integrator')[1])
        for k in range(N*M):
          self.assertAlmostEqual(x2sol[k+1], x1sol[k])
          self.assertAlmostEqual(x1sol[k+1], (1 - x2sol[k]**2) * x1sol[k] - x2sol[k] + usol[k])

      
    def test_DT(self):

      for system in ['set_next','set_der']:
          ocp = Ocp(T=10)

          x = ocp.state()

          u = ocp.control()
          
          getattr(ocp,system)(x, u)

          ocp.add_objective(ocp.sum(u**2))

          ocp.subject_to(-1 <= (u <= 1))

          ocp.subject_to(ocp.at_t0(x) == 0)
          ocp.subject_to(ocp.at_tf(x) == 1)
          
          for M in [1,2,3]:

              ocp.solver('ipopt')
              ocp.method(MultipleShooting(N=6,M=M,grid=GeometricGrid(2)))
              sol = ocp.solve()

              [ts,DT_controls] = sol.sample(ocp.DT_control,grid='control')
              np.testing.assert_allclose(np.diff(ts),DT_controls[:-1])
              self.assertEqual(DT_controls[-1],DT_controls[-2])
              
              [_,DT_controls] = sol.sample(ocp.DT_control,grid='integrator')
              np.testing.assert_allclose(np.array(kron(np.diff(ts),[1]*M)).squeeze(),DT_controls[:-1])

              [ts,DTs] = sol.sample(ocp.DT,grid='integrator')
              np.testing.assert_allclose(np.diff(ts),DTs[:-1])
              self.assertEqual(DTs[-1],DTs[-2])

              [ts,DTs_control] = sol.sample(ocp.DT,grid='control')
              np.testing.assert_allclose(DTs[::M],DTs_control)

    def test_DT_system(self):
    
      for dt in ["DT_control","DT",None]:

          for system in ['set_next','set_der']:
              ocp = Ocp(T=10)

              x = ocp.state()

              u = ocp.control()
              
              getattr(ocp,system)(x, (1 if dt is None else getattr(ocp,dt))*u)

              ocp.subject_to(-1 <= (u <= 1))

              ocp.subject_to(ocp.at_t0(x) == 0)
              ocp.add_objective(-ocp.at_tf(x))
              
              for N in [5,10]:
                  for M in [1,2]:

                      ocp.solver('ipopt',{"ipopt.print_level":0,"print_time":False})
                      ocp.method(MultipleShooting(N=N,M=M))
                      if system=="set_der" and dt is not None:
                          with self.assertRaisesRegex(Exception, dt):
                            sol = ocp.solve()
                      else:
                        sol = ocp.solve()
                        xf = sol.value(ocp.at_tf(x))
                        if system=='set_next':
                            if dt == "DT_control":
                                DT_control = 10/N
                                self.assertAlmostEqual(xf, DT_control)
                            elif dt == "DT":
                                DT = 10/N/M
                                self.assertAlmostEqual(xf, DT)
                            else:
                                self.assertAlmostEqual(xf, 1)

    def test_initial(self):
        for t0_stage, t0_sol_stage in [(None, 0), (-1, -1), (FreeTime(-1), -1)]:
            for T_stage, T_sol_stage in [(None, 2), (2, 2), (FreeTime(1), 2)]:
                kwargs = {}
                if t0_stage is not None:
                    kwargs["t0"] = t0_stage
                if T_stage is not None:
                    kwargs["T"] = T_stage
                stage = Stage(**kwargs)

                p = stage.state()
                v = stage.state()
                u = stage.control()

                stage.set_der(p, v)
                stage.set_der(v, u)

                stage.subject_to(u <= 1)
                stage.subject_to(-1 <= u)

                #stage.set_initial(p, stage.t)

                stage.add_objective(stage.tf)
                stage.method(MultipleShooting(N=2))

                for t0, t0_sol in ([] if t0_stage is None else [(None, t0_sol_stage)]) + [(-1, -1), (FreeTime(-1), -1)]:
                    for T, T_sol in ([] if T_stage is None else [(None, T_sol_stage)]) + [(2, 2), (FreeTime(1), 2)]:
                        ocp = Ocp()

                        kwargs = {}
                        if t0 is not None:
                            kwargs["t0"] = t0
                        if T is not None:
                            kwargs["T"] = T
                        mystage = ocp.stage(stage, **kwargs)

                        if isinstance(t0,FreeTime) or (t0 is None and isinstance(t0_stage,FreeTime)):
                          print("stage", mystage.t0)
                          ocp.subject_to(mystage.t0 >= t0_sol)

                        ocp.solver('ipopt',{"ipopt.max_iter": 0})
                        mystage.set_initial(p, mystage.t)
                        sol = ocp.solve_limited()


                        t0_num = sol(mystage).value(mystage.t0)
                        T_num = sol(mystage).value(mystage.T)

                        tolerance = 1e-6

                        ps_ref = np.linspace(t0_num, t0_num+T_num, 2+1)
                        ts, ps = sol(mystage).sample(p, grid='control')
                        np.testing.assert_allclose(ps, ps_ref, atol=tolerance)

    def test_stage_cloning_t0_T(self):
        for t0_stage, t0_sol_stage in [(None, 0), (-1, -1), (FreeTime(-1), -1)]:
            for T_stage, T_sol_stage in [(None, 2), (2, 2), (FreeTime(1), 2)]:
                kwargs = {}
                if t0_stage is not None:
                    kwargs["t0"] = t0_stage
                if T_stage is not None:
                    kwargs["T"] = T_stage
                stage = Stage(**kwargs)

                p = stage.state()
                v = stage.state()
                u = stage.control()

                stage.set_der(p, v)
                stage.set_der(v, u)

                stage.subject_to(u <= 1)
                stage.subject_to(-1 <= u)

                stage.add_objective(stage.tf)
                stage.subject_to(stage.at_t0(p) == 0)
                stage.subject_to(stage.at_t0(v) == 0)
                stage.subject_to(stage.at_tf(p) == 1)
                stage.subject_to(stage.at_tf(v) == 0)
                stage.method(MultipleShooting(N=2))

                for t0, t0_sol in ([] if t0_stage is None else [(None, t0_sol_stage)]) + [(-1, -1), (FreeTime(-1), -1)]:
                    for T, T_sol in ([] if T_stage is None else [(None, T_sol_stage)]) + [(2, 2), (FreeTime(1), 2)]:
                        ocp = Ocp()

                        kwargs = {}
                        if t0 is not None:
                            kwargs["t0"] = t0
                        if T is not None:
                            kwargs["T"] = T
                        mystage = ocp.stage(stage, **kwargs)

                        if isinstance(t0,FreeTime) or (t0 is None and isinstance(t0_stage,FreeTime)):
                          print("stage", mystage.t0)
                          ocp.subject_to(mystage.t0 >= t0_sol)

                        ocp.solver('ipopt')

                        sol = ocp.solve()

                        tolerance = 1e-6

                        ts, ps = sol(mystage).sample(p, grid='integrator', refine=10)

                        ps_ref = np.hstack(((0.5*np.linspace(0,1, 10+1)**2)[:-1],np.linspace(0.5,1.5,10+1)-0.5*np.linspace(0,1, 10+1)**2)) 
                        np.testing.assert_allclose(ps, ps_ref, atol=tolerance)

                        ts_ref = t0_sol + np.linspace(0, 2, 10*2+1)

                        ts, vs = sol(mystage).sample(v, grid='integrator', refine=10)
                        np.testing.assert_allclose(ts, ts_ref, atol=tolerance)

                        vs_ref = np.hstack((np.linspace(0,1, 10+1)[:-1],np.linspace(1,0, 10+1))) 
                        np.testing.assert_allclose(vs, vs_ref, atol=tolerance)


                        u_ref = np.array([1.0]*10+[-1.0]*11)
                        ts, us = sol(mystage).sample(u, grid='integrator', refine=10)
                        np.testing.assert_allclose(us, u_ref, atol=tolerance)

if __name__ == '__main__':
    unittest.main()



# -*- coding: future_fstrings -*-
#
# Copyright (c) The acados authors.
#
# This file is part of acados.
#
# The 2-Clause BSD License
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
# this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.;
#


cdef extern from "acados/ocp_nlp/ocp_nlp_common.h":
    ctypedef struct ocp_nlp_config:
        pass

    ctypedef struct ocp_nlp_dims:
        pass

    ctypedef struct ocp_nlp_in:
        pass

    ctypedef struct ocp_nlp_out:
        pass


cdef extern from "acados_c/ocp_nlp_interface.h":
    ctypedef enum ocp_nlp_solver_t:
        pass

    ctypedef enum ocp_nlp_cost_t:
        pass

    ctypedef enum ocp_nlp_dynamics_t:
        pass

    ctypedef enum ocp_nlp_constraints_t:
        pass

    ctypedef enum ocp_nlp_reg_t:
        pass

    ctypedef struct ocp_nlp_plan:
        pass

    ctypedef struct ocp_nlp_solver:
        pass

    int ocp_nlp_cost_model_set(ocp_nlp_config *config, ocp_nlp_dims *dims, ocp_nlp_in *in_,
        int start_stage, const char *field, void *value)
    int ocp_nlp_constraints_model_set(ocp_nlp_config *config, ocp_nlp_dims *dims,
        ocp_nlp_in *in_, int stage, const char *field, void *value)

    # out
    void ocp_nlp_out_set(ocp_nlp_config *config, ocp_nlp_dims *dims, ocp_nlp_out *out,
        int stage, const char *field, void *value)
    void ocp_nlp_out_get(ocp_nlp_config *config, ocp_nlp_dims *dims, ocp_nlp_out *out,
        int stage, const char *field, void *value)
    void ocp_nlp_get_at_stage(ocp_nlp_solver *solver, int stage, const char *field, void *value)
    int ocp_nlp_dims_get_from_attr(ocp_nlp_config *config, ocp_nlp_dims *dims, ocp_nlp_out *out,
        int stage, const char *field)
    void ocp_nlp_constraint_dims_get_from_attr(ocp_nlp_config *config, ocp_nlp_dims *dims, ocp_nlp_out *out,
        int stage, const char *field, int *dims_out)
    void ocp_nlp_cost_dims_get_from_attr(ocp_nlp_config *config, ocp_nlp_dims *dims, ocp_nlp_out *out,
        int stage, const char *field, int *dims_out)
    void ocp_nlp_qp_dims_get_from_attr(ocp_nlp_config *config, ocp_nlp_dims *dims, ocp_nlp_out *out,
        int stage, const char *field, int *dims_out)

    # in
    void ocp_nlp_in_set(ocp_nlp_config *config, ocp_nlp_dims *dims, ocp_nlp_in *nlp_in,
        int stage, const char *field, void *value)
    void ocp_nlp_in_get(ocp_nlp_config *config, ocp_nlp_dims *dims, ocp_nlp_in *nlp_in,
        int stage, const char *field, void *value)

    # opts
    void ocp_nlp_solver_opts_set(ocp_nlp_config *config, void *opts_, const char *field, void* value)

    # solver
    void ocp_nlp_eval_residuals(ocp_nlp_solver *solver, ocp_nlp_in *nlp_in, ocp_nlp_out *nlp_out)
    void ocp_nlp_eval_param_sens(ocp_nlp_solver *solver, char *field, int stage, int index, ocp_nlp_out *sens_nlp_out)
    void ocp_nlp_eval_cost(ocp_nlp_solver *solver, ocp_nlp_in *nlp_in_, ocp_nlp_out *nlp_out)
    void ocp_nlp_eval_params_jac(ocp_nlp_solver *solver, ocp_nlp_in *nlp_in_, ocp_nlp_out *nlp_out)
    void ocp_nlp_eval_lagrange_grad_p(ocp_nlp_solver *solver, ocp_nlp_in *nlp_in_, const char *field, void* value)
    # get/set
    void ocp_nlp_get(ocp_nlp_solver *solver, const char *field, void *return_value_)
    void ocp_nlp_set(ocp_nlp_solver *solver, int stage, const char *field, void *value)



#! /usr/bin/python

import subprocess
import sys
import json
from jinja2 import Template

class RoutinesLibrary:
    def __init__(self, routines="routines.json"):

        with open(routines) as f:
            self.routines = json.load(f)


    def generate_docs(self, output_file="routines_list.md", template="routines_list.md.tpl"):

        with open(template) as f:
            template_txt = f.read()

        self.preprocess()
        template = Template(template_txt)
        output = template.render(routines=self.routines)
        with open(output_file, "w") as f:
            f.write(output)
        print(output)

    def preprocess(self):
        for rtn_class_name, rtn_class in self.routines.items():
            if rtn_class_name=="aux": self.build_subclasses("aux", skip=["mem"])
            self.add_precision(rtn_class_name)

    def build_subclasses(self, rtn_class, skip=[]):
        built_class = {}
        for rtn_subclass_name, rtn_subclass in self.routines[rtn_class].items():

            if rtn_subclass_name in skip:
                built_class.update(rtn_subclass)
                continue
            for rtn_name, rtn_info in rtn_subclass.items():
                built_class[f"{rtn_subclass_name}{rtn_name}"] = rtn_info

        self.routines[rtn_class] = built_class

    def add_precision(self, rtn_class):
        built_class = {}
        for rtn_name, rtn_info in self.routines[rtn_class].items():

            if "{PREC}" not in rtn_name and "STR" not in rtn_name:
                # default naming
                routine_name = f"blasfeo_(s|d){rtn_name}".replace("_", "\_")
            else:
                # templated naming
                routine_name="blasfeo_"+rtn_name.format(
                    PREC="(s|d)",
                    STRVEC="(s|d)vec",
                    STRMAT="(s|d)mat")
                routine_name = routine_name.replace("_", "\_")

            built_class[routine_name] = rtn_info


        self.routines[rtn_class] = built_class


if __name__ == "__main__":

    # generate recipes
    # test set to be run in the given excution of the script
    routines_lib = RoutinesLibrary()
    #  print(json.dumps(cookbook.recipe, indent=4))
    routines_lib.generate_docs()



#
#     This file is part of rockit.
#
#     rockit -- Rapid Optimal Control Kit
#     Copyright (C) 2019 MECO, KU Leuven. All rights reserved.
#
#     Rockit is free software; you can redistribute it and/or
#     modify it under the terms of the GNU Lesser General Public
#     License as published by the Free Software Foundation; either
#     version 3 of the License, or (at your option) any later version.
#
#     Rockit is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#     Lesser General Public License for more details.
#
#     You should have received a copy of the GNU Lesser General Public
#     License along with CasADi; if not, write to the Free Software
#     Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
#
#

from casadi import Opti, jacobian, dot, hessian, symvar, evalf, veccat, DM, vertcat, is_equal
import casadi
import numpy as np
from .casadi_helpers import get_meta, merge_meta, single_stacktrace, MX
from .solution import OcpSolution
from .freetime import FreeTime

class DirectMethod:
    """
    Base class for 'direct' solution methods for Optimal Control Problems:
      'first discretize, then optimize'
    """
    def __init__(self):
        self._solver = None
        self._solver_options = None
        self._callback = None
        self.artifacts = []
        self.clean()

    def clean(self):
        self.V = None
        self.P = []

    def jacobian(self, with_label=False):
        J = jacobian(self.opti.g, self.opti.x).sparsity()
        if with_label:
            return J, "Constraint Jacobian: " + J.dim(True)
        else:
            return J

    def hessian(self, with_label=False):
        lag = self.opti.f + dot(self.opti.lam_g, self.opti.g)
        H = hessian(lag, self.opti.x)[0].sparsity()
        if with_label:
            return H, "Lagrange Hessian: " + H.dim(True)
        else:
            return H

    def spy_jacobian(self):
        import matplotlib.pylab as plt
        J, title = self.jacobian(with_label=True)
        plt.spy(np.array(J))
        plt.title(title)

    def spy_hessian(self):
        import matplotlib.pylab as plt
        lag = self.opti.f + dot(self.opti.lam_g, self.opti.g)
        H, title = self.hessian(with_label=True)
        plt.spy(np.array(H))
        plt.title(title)
    
    def inherit(self, template):
        if template and template._solver is not None:
            self._solver = template._solver
        if template and template._solver_options is not None:
            self._solver_options = template._solver_options
        if template and template._callback is not None:
            self._callback = template._callback

    def eval(self, stage, expr):
        return self.eval_top(stage, expr)

    def eval_top(self, stage, expr):
        return substitute(MX(expr),veccat(*(stage.variables[""]+stage.parameters[""])),vertcat(self.V,veccat(*self.P)))

    def add_variables(self, stage, opti):
        V = []
        for v in stage.variables['']:
            V.append(opti.variable(v.shape[0], v.shape[1], scale=stage._scale[v], domain=stage._catalog[v]['domain']))
        self.V = veccat(*V)

    def add_parameters(self, stage, opti):
        self.P = []
        for p in stage.parameters['']:
            self.P.append(opti.parameter(p.shape[0], p.shape[1]))

    def untranscribe(self, stage, phase=1,**kwargs):
        self.clean()

    def untranscribe_placeholders(self, phase, stage):
        pass

    def main_untranscribe(self,stage, phase=1, **kwargs):
        self.opti = None

    def main_transcribe(self, stage, phase=1, **kwargs):
        if phase==0: return
        if phase==1:
            self.opti = OptiWrapper(stage)
            if self._callback:
                self.opti.callback(self._callback)
            if self.solver is not None:
                if self._solver is None:
                    raise Exception("You forgot to declare a solver. Use e.g. ocp.solver('ipopt').")
                self.opti.solver(self._solver, self._solver_options)
        if phase==2:
            self.opti.transcribe_placeholders(phase, kwargs["placeholders"])

    def transcribe(self, stage, phase=1, **kwargs):
        if stage.nx>0 or stage.nu>0:
            raise Exception("You forgot to declare a method. Use e.g. ocp.method(MultipleShooting(N=3)).")
        if phase==0: return
        if phase>1: return
        self.add_variables(stage, self.opti)
        self.add_parameters(stage, self.opti)

        for c, m, _ in stage._constraints["point"]:
            self.opti.subject_to(self.eval_top(stage, c), meta = m)
        self.opti.add_objective(self.eval_top(stage, stage._objective))
        self.set_initial(stage, self.opti, stage._initial)
        self.set_parameter(stage, self.opti)

    def set_initial(self, stage, master, initial):
        opti = master.opti if hasattr(master, 'opti') else master
        opti.cache_advanced()
        for var, expr in initial.items():
            opti_initial = opti.initial()
            target = self.eval_top(stage, var)
            value = DM(opti.debug.value(self.eval_top(stage, expr), opti_initial)) # HOT line
            opti.set_initial(target, value, cache_advanced=True)

    def set_parameter(self, stage, opti):
        for i, p in enumerate(stage.parameters['']):
            opti.set_value(self.P[i], stage._param_value(p))

    def set_value(self, stage, master, parameter, value):
        opti = master.opti if hasattr(master, 'opti') else master
        found = False
        for i, p in enumerate(stage.parameters['']):
            if is_equal(parameter, p):
                found = True
                opti.set_value(self.P[i], value)
        assert found, "You attempted to set the value of a non-parameter."

    def transcribe_placeholders(self, phase, stage, placeholders):
        pass

    def non_converged_solution(self, stage):
        if not hasattr(self, 'opti'):
            raise Exception("You forgot to solve first. To avoid your script halting, use a try-catch block.")
        return OcpSolution(self.opti.non_converged_solution, stage)

    def solve(self, stage):
        return OcpSolution(self.opti.solve(), stage)

    def solve_limited(self, stage):
        return OcpSolution(self.opti.solve_limited(), stage)

    def callback(self, stage, fun):
        self._callback = lambda iter : fun(iter, OcpSolution(self.opti.non_converged_solution, stage))

    @property
    def debug(self):
        self.opti.debug

    def solver(self, solver, solver_options={}):
        self._solver = solver
        self._solver_options = solver_options

    def show_infeasibilities(self, *args):
        self.opti.debug.show_infeasibilities(*args)

    def initial_value(self, stage, expr):
        return self.opti.value(expr, self.opti.initial())

    @property
    def gist(self):
        """Obtain an expression packing all information needed to obtain value/sample

        The composition of this array may vary between rockit versions

        Returns
        -------
        :obj:`~casadi.MX` column vector

        """
        return vertcat(self.opti.x, self.opti.p)

    def to_function(self, stage, name, args, results, *margs):
        return self.opti.to_function(name, [stage.value(a) for a in args], results, *margs)

    def fill_placeholders_integral(self, phase, stage, expr, *args):
        if phase==1:
            I = stage.state(quad=True)
            stage.set_der(I, expr)
            return stage.at_tf(I)

    def fill_placeholders_T(self, phase, stage, expr, *args):
        if phase==1:
            if isinstance(stage._T, FreeTime):
                init = stage._T.T_init
                stage.set_T(stage.variable())
                stage.subject_to(stage._T>=0)
                stage.set_initial(stage._T, init,priority=True)
                return stage._T
            else:
                return stage._T
        return self.eval(stage, expr)

    def fill_placeholders_t0(self, phase, stage, expr, *args):
        if phase==1:
            if isinstance(stage._t0, FreeTime):
                init = stage._t0.T_init
                stage.set_t0(stage.variable())
                stage.set_initial(stage._t0, init,priority=True)
                return stage._t0
            else:
                return stage._t0
        return self.eval(stage, expr)

    def fill_placeholders_t(self, phase, stage, expr, *args):
        return None
    
    def fill_placeholders_DT(self, phase, stage, expr, *args):
        return None

    def fill_placeholders_DT_control(self, phase, stage, expr, *args):
        return None


from casadi import substitute

class OptiWrapper(Opti):
    def __init__(self, ocp):
        self.ocp = ocp
        Opti.__init__(self)
        self.initial_keys = []
        self.initial_values = []
        self.constraints = []
        self.objective = 0

    def subject_to(self, expr=None, scale=1, meta=None):
        meta = merge_meta(meta, get_meta())
        if expr is None:
            self.constraints = []
        else:
            if isinstance(expr,MX) and expr.is_constant():
                if np.all(np.array(evalf(expr)).squeeze()==1):
                    return
                else:
                    raise Exception("You have a constraint that is never statisfied.")
            self.constraints.append((expr, scale, meta))

    def add_objective(self, expr):
        self.objective = self.objective + expr

    def clear_objective(self):
        self.objective = 0

    def callback(self,fun):
        Opti.callback(self, fun)

    def initial(self):
        return [e for e in Opti.initial(self) if e.dep(0).is_symbolic() or e.dep(1).is_symbolic()]+Opti.value_parameters(self)

    @property
    def non_converged_solution(self):
        return OptiSolWrapper(self, self.debug)

    def variable(self,n=1,m=1, scale=1,domain='real'):
        if n==0 or m==0:
            return MX(n, m)
        else:
            v = Opti.variable(self,n, m)
            if hasattr(Opti,'set_domain'):
                Opti.set_domain(self, v, domain)
            else:
                if domain!='real': raise Exception("This version of CasADi Opti stack does not support set_domain.")
            return scale*v

    def cache_advanced(self):
        self._advanced_cache = self.advanced

    def set_initial(self, key, value, cache_advanced=False):
        a = set([hash(e) for e in (self._advanced_cache if cache_advanced else self.advanced).symvar()])
        b = set([hash(e) for e in symvar(key)])
        if len(a | b)==len(a):
            Opti.set_initial(self, key, value) # set_initial logic in direct_collocation needs this
        else:
            self.initial_keys.append(key)
            self.initial_values.append(value)

    def transcribe_placeholders(self,phase,placeholders):
        opti_advanced = self.advanced
        Opti.subject_to(self)
        n_constr = len(self.constraints)
        res = placeholders([c[0] for c in self.constraints] + [self.objective]+self.initial_keys)
        for c, scale, meta in zip(res[:n_constr], [c[1] for c in self.constraints], [c[2] for c in self.constraints]):
            try:
                if MX(c).is_constant() and MX(c).is_one():
                    continue
                if not MX(scale).is_one():
                    mc = opti_advanced.canon_expr(c) # canon_expr should have a static counterpart
                    if mc.type in [casadi.OPTI_INEQUALITY, casadi.OPTI_GENERIC_INEQUALITY, casadi.OPTI_DOUBLE_INEQUALITY]:
                        print(mc.lb,mc.canon,mc.ub)
                        lb = mc.lb/scale
                        canon = mc.canon/scale
                        ub = mc.ub/scale
                        # Check for infinities
                        try:
                            lb_inf = np.all(np.array(evalf(lb)==-np.inf))
                        except:
                            lb_inf = False
                        try:
                            ub_inf = np.all(np.array(evalf(ub)==np.inf))
                        except:
                            ub_inf = False
                        if lb_inf:
                            c = canon <= ub
                        elif ub_inf:
                            c = lb <= canon
                        else:     
                            c = lb <= (canon <= ub)
                    if mc.type in [casadi.OPTI_EQUALITY, casadi.OPTI_GENERIC_EQUALITY]:
                        lb = mc.lb/scale
                        canon = mc.canon/scale
                        c = lb==canon
                Opti.subject_to(self,c)
            except Exception as e:
                print(meta,c)
                raise e
            self.update_user_dict(c, single_stacktrace(meta))
        Opti.minimize(self,res[n_constr])
        for k_orig, k, v in zip(self.initial_keys,res[n_constr+1:],self.initial_values):
            Opti.set_initial(self, k, v)


    def solve(self):
        return OptiSolWrapper(self, Opti.solve(self))

class OptiSolWrapper:
    def __init__(self, opti_wrapper, sol):
        self.opti_wrapper = opti_wrapper
        self.sol = sol

    def value(self, expr, *args,**kwargs):
        placeholders = self.opti_wrapper.ocp.placeholders_transcribed
        return self.sol.value(placeholders(expr), *args, **kwargs)

    def stats(self):
        return self.sol.stats()


from pylab import *
import unittest

from rockit import Ocp, DirectMethod, MultipleShooting, FreeTime, Stage
import numpy as np
from casadi import kron, DM

from .method import GrampcMethod

from ...casadi_helpers import AutoBrancher

class GrampcTests(unittest.TestCase):
    pass

if __name__ == '__main__':
    unittest.main()



#
#     This file is part of rockit.
#
#     rockit -- Rapid Optimal Control Kit
#     Copyright (C) 2019 MECO, KU Leuven. All rights reserved.
#
#     Rockit is free software; you can redistribute it and/or
#     modify it under the terms of the GNU Lesser General Public
#     License as published by the Free Software Foundation; either
#     version 3 of the License, or (at your option) any later version.
#
#     Rockit is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#     Lesser General Public License for more details.
#
#     You should have received a copy of the GNU Lesser General Public
#     License along with CasADi; if not, write to the Free Software
#     Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
#
#

from .sampling_method import SamplingMethod
from casadi import sumsqr, horzcat, linspace, substitute, MX, evalf, vcat, horzsplit, veccat, DM, repmat, vvcat, vcat, vec
import numpy as np

class SingleShooting(SamplingMethod):
    def __init__(self, **kwargs):
        SamplingMethod.__init__(self, **kwargs)

    def add_parameter(self, stage, opti):
        SamplingMethod.add_parameter(self, stage, opti)
        self.Z0 = []
        es = []
        for s in stage.algebraics:
            e = opti.parameter(s.numel())
            opti.set_value(e, 0)
            es.append(e)
        self.Z0.append(vcat(es))
        self.Z0 += [None]*(self.N-1)

    def add_variables(self,stage,opti):
        scale_x = stage._scale_x
        scale_u = stage._scale_u

        # We are creating variables in a special order such that the resulting constraint Jacobian
        # is block-sparse
        self.X.append(vcat([opti.variable(s.numel(), scale=vec(stage._scale[s])) for s in stage.states]))
        self.add_variables_V(stage, opti)
        self.Q.append(DM.zeros(stage.nxq))

        for k in range(self.N):
            self.U.append(vcat([opti.variable(s.numel(), scale=vec(stage._scale[s]), domain=stage._catalog[s]['domain']) for s in stage.controls]) if stage.nu>0 else MX(0,1))
            self.X.append(None)
            self.Q.append(None)
            self.add_variables_V_control(stage, opti, k)

        self.add_variables_V_control_finalize(stage, opti)

    def add_constraints(self,stage,opti):
        self.add_constraints_before(stage, opti)
        # Obtain the discretised system
        F = self.discrete_system(stage)

        # we only save polynomal coeffs for runge-kutta4
        if F.numel_out("poly_coeff")==0:
            self.poly_coeff = None
        if F.numel_out("poly_coeff_z")==0:
            self.poly_coeff_z = None
        if F.numel_out("poly_coeff_q")==0:
            self.poly_coeff_q = None

        Z0 = self.Z0[0]

        self.q = DM.zeros(stage.nxq)
        FFs = []
        self.xqk.append(DM.zeros(stage.nxq))
        # Fill in Z variables up-front, since they might be needed in constraints with ocp.next
        for k in range(self.N):
            FF = F(x0=self.X[k], u=self.U[k], t0=self.control_grid[k],
                   T=self.control_grid[k + 1] - self.control_grid[k], p=self.get_p_sys(stage, k), z0=Z0)
            self.X[k+1] = FF["xf"]
            poly_coeff_temp = FF["poly_coeff"]
            poly_coeff_z_temp = FF["poly_coeff_z"]
            poly_coeff_q_temp = FF["poly_coeff_q"]
            xk_temp = FF["Xi"]
            xqk_temp = self.q + FF["Qi"]
            zk_temp = FF["Zi"]
            # we cannot return a list from a casadi function
            self.xk.extend([xk_temp[:, i] for i in range(self.M)])
            self.xqk.extend([xqk_temp[:, i] for i in range(self.M)])
            self.zk.extend([zk_temp[:, i] for i in range(self.M)])
            if k==0:
                self.Z.append(zk_temp[:, 0])
            self.Z.append(FF["zf"])
            if self.poly_coeff is not None:
                self.poly_coeff.extend(horzsplit(poly_coeff_temp, poly_coeff_temp.shape[1]//self.M))
            if self.poly_coeff_q is not None:
                self.poly_coeff_q.extend(horzsplit(poly_coeff_q_temp, poly_coeff_q_temp.shape[1]//self.M))
            if self.poly_coeff_z is not None:
                self.poly_coeff_z.extend(horzsplit(poly_coeff_z_temp, poly_coeff_z_temp.shape[1]//self.M))
            FFs.append(FF)
            self.q = self.q + FF["qf"]
            self.Q[k+1] = self.q

        self.xk.append(self.X[-1])
        self.zk.append(self.zk[-1])

        for k in range(self.N):
            FF = FFs[k]

            for l in range(self.M):
                for c, meta, args in stage._constraints["integrator"]:
                    if k==0 and l==0 and not args["include_first"]: continue
                    opti.subject_to(self.eval_at_integrator(stage, c, k, l), scale=args["scale"], meta=meta)
                for c, meta, _ in stage._constraints["inf"]:
                    self.add_inf_constraints(stage, opti, c, k, l, meta)

            for c, meta, args in stage._constraints["control"]:  # for each constraint expression
                if k==0 and not args["include_first"]: continue
                try:
                    opti.subject_to(self.eval_at_control(stage, c, k), scale=args["scale"], meta=meta)
                except IndexError:
                    pass # Can be caused by ocp.offset -> drop constraint

        for c, meta, args in stage._constraints["control"]+stage._constraints["integrator"]:  # for each constraint expression
            if not args["include_last"]: continue
            # Add it to the optimizer, but first make x,u concrete.
            try:
                opti.subject_to(self.eval_at_control(stage, c, -1), scale=args["scale"], meta=meta)
            except IndexError:
                pass # Can be caused by ocp.offset -> drop constraint



#
# Copyright (c) The acados authors.
#
# This file is part of acados.
#
# The 2-Clause BSD License
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
# this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.;
#


import shutil
import matplotlib

def latexify_plot() -> None:
    text_usetex = True if shutil.which('latex') else False
    params = {
            'text.latex.preamble': r"\usepackage{gensymb} \usepackage{amsmath}",
            'axes.labelsize': 12,
            'axes.titlesize': 12,
            'legend.fontsize': 12,
            'xtick.labelsize': 12,
            'ytick.labelsize': 12,
            'text.usetex': text_usetex,
            'font.family': 'serif'
    }

    matplotlib.rcParams.update(params)
    return


# -*- coding: future_fstrings -*-
#
# Copyright (c) The acados authors.
#
# This file is part of acados.
#
# The 2-Clause BSD License
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
# this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.;
#


cdef extern from "acados/sim/sim_common.h":
    ctypedef struct sim_config:
        pass

    ctypedef struct sim_opts:
        pass

    ctypedef struct sim_in:
        pass

    ctypedef struct sim_out:
        pass


cdef extern from "acados_c/sim_interface.h":

    ctypedef struct sim_plan:
        pass

    ctypedef struct sim_solver:
        pass

    # out
    void sim_out_get(sim_config *config, void *dims, sim_out *out, const char *field, void *value)
    int sim_dims_get_from_attr(sim_config *config, void *dims, const char *field, void *dims_data)

    # opts
    void sim_opts_set(sim_config *config, void *opts_, const char *field, void *value)

    # get/set
    void sim_in_set(sim_config *config, void *dims, sim_in *sim_in, const char *field, void *value)
    void sim_solver_set(sim_solver *solver, const char *field, void *value)


# -*- coding: future_fstrings -*-
#
# Copyright (c) The acados authors.
#
# This file is part of acados.
#
# The 2-Clause BSD License
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
# this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.;
#

from typing import Union, List
import numpy as np
import casadi as ca
from copy import deepcopy

import os, json

from .acados_model import AcadosModel
from .acados_dims import AcadosOcpDims
from .acados_ocp_cost import AcadosOcpCost
from .acados_ocp_constraints import AcadosOcpConstraints
from .acados_ocp_options import AcadosOcpOptions, INTEGRATOR_TYPES, COLLOCATION_TYPES, COST_DISCRETIZATION_TYPES
from .acados_ocp import AcadosOcp
from .casadi_function_generation import GenerateContext
from .utils import make_object_json_dumpable, get_acados_path, format_class_dict, get_shared_lib_ext, render_template, is_empty


def find_non_default_fields_of_obj(obj: Union[AcadosOcpCost, AcadosOcpConstraints, AcadosOcpOptions], stage_type='all') -> list:

    all_fields = [field for field in dir(obj) if not field.startswith("_")]

    # remove special properties which are translated to other fields
    if isinstance(obj, AcadosOcpConstraints):
        all_fields.remove('x0')
        all_fields = [field for field in all_fields if not field.startswith("J")]

    if isinstance(obj, AcadosOcpOptions):
        all_fields.remove('qp_tol')
        all_fields.remove('tol')

    all_fields = [field for field in all_fields if not callable(getattr(obj, field))]

    if stage_type == 'all':
        pass
    elif stage_type == 'initial':
        all_fields = [field for field in all_fields if field.endswith("_0")]
    elif stage_type == 'terminal':
        all_fields = [field for field in all_fields if field.endswith("_e")]
    else:
        raise Exception(f"stage_type {stage_type} not supported.")

    obj_type = type(obj)
    dummy_obj = obj_type()
    nondefault_fields = []
    for field in all_fields:
        val = getattr(obj, field)
        default_val = getattr(dummy_obj, field)
        if not isinstance(val, type(default_val)):
            nondefault_fields.append(field)

        elif isinstance(val, np.ndarray):
            if not np.array_equal(val, default_val):
                nondefault_fields.append(field)

        elif val != default_val:
            nondefault_fields.append(field)

    return nondefault_fields


class AcadosMultiphaseOptions:
    """
    Class containing options that might be different for each phase.

    All of the fields can be either None, then the corresponding value from ocp.solver_options is used,
    or a list of length n_phases describing the value for this option at each phase.

    - integrator_type: list of strings, must be in ["ERK", "IRK", "GNSF", "DISCRETE", "LIFTED_IRK"]
    - collocation_type: list of strings, must be in ["GAUSS_RADAU_IIA", "GAUSS_LEGENDRE", "EXPLICIT_RUNGE_KUTTA"]
    - cost_discretization: list of strings, must be in ["EULER", "INTEGRATOR"]
    """
    def __init__(self):
        self.integrator_type = None
        self.collocation_type = None
        self.cost_discretization = None

    def make_consistent(self, opts: AcadosOcpOptions, n_phases: int) -> None:
        for field, variants in zip(['integrator_type', 'collocation_type', 'cost_discretization'],
                                [INTEGRATOR_TYPES, COLLOCATION_TYPES, COST_DISCRETIZATION_TYPES]
                                ):
            if getattr(self, field) is None:
                # non varying field, use value from ocp opts
                setattr(self, field, [getattr(opts, field) for _ in range(n_phases)])
            if not isinstance(getattr(self, field), list):
                raise Exception(f'AcadosMultiphaseOptions.{field} must be a list, got {getattr(self, field)}.')
            if len(getattr(self, field)) != n_phases:
                raise Exception(f'AcadosMultiphaseOptions.{field} must be a list of length n_phases, got {getattr(self, field)}.')
            if not all([item in variants for item in getattr(self, field)]):
                raise Exception(f'AcadosMultiphaseOptions.{field} must be a list of strings in {variants}, got {getattr(self, field)}.')


class AcadosMultiphaseOcp:
    """
    Class containing the description of an optimal control problem with multiple phases.
    This object can be used to create an :py:class:`acados_template.acados_ocp_solver.AcadosOcpSolver`.

    Initial cost and constraints are defined by the first phase, terminal cost and constraints by the last phase.
    All other phases are treated as intermediate phases, where only dynamics and path cost and constraints are used.

    Solver options are shared between all phases. Options that can vary phase-wise must be set via self.mocp_opts of type :py:class:`acados_template.acados_multiphase_ocp.AcadosMultiphaseOptions`.

    :param N_list: list containing the number of shooting intervals for each phase
    """
    def __init__(self, N_list: list):

        if not isinstance(N_list, list) or len(N_list) < 1:
            raise Exception("N_list must be a list of integers.")
        if any([not isinstance(N, int) for N in N_list]):
            raise Exception("N_list must be a list of integers.")

        n_phases = len(N_list)

        self.n_phases = n_phases
        self.N_list = N_list

        self.name = 'multiphase_ocp'
        self.model = [AcadosModel() for _ in range(n_phases)]
        """Model definitions, type :py:class:`acados_template.acados_model.AcadosModel`"""
        self.cost = [AcadosOcpCost() for _ in range(n_phases)]
        """Cost definitions, type :py:class:`acados_template.acados_ocp_cost.AcadosOcpCost`"""
        self.constraints = [AcadosOcpConstraints() for _ in range(n_phases)]
        """Constraints definitions, type :py:class:`acados_template.acados_ocp_constraints.AcadosOcpConstraints`"""

        self.phases_dims = [AcadosOcpDims() for _ in range(n_phases)]

        self.dummy_ocp_list: List[AcadosOcp] = []

        # NOTE: this is the same for AcadosOcp
        self.solver_options = AcadosOcpOptions()
        """Solver Options, type :py:class:`acados_template.acados_ocp_options.AcadosOcpOptions`"""
        self.mocp_opts = AcadosMultiphaseOptions()
        """Phase-wise varying solver Options, type :py:class:`acados_template.acados_multiphase_ocp.AcadosMultiphaseOptions`"""

        acados_path = get_acados_path()

        self.acados_include_path = os.path.join(acados_path, 'include').replace(os.sep, '/') # the replace part is important on Windows for CMake
        """Path to acados include directory (set automatically), type: `string`"""
        self.acados_lib_path = os.path.join(acados_path, 'lib').replace(os.sep, '/') # the replace part is important on Windows for CMake
        """Path to where acados library is located, type: `string`"""
        self.shared_lib_ext = get_shared_lib_ext()

        # get cython paths
        from sysconfig import get_paths
        self.cython_include_dirs = [np.get_include(), get_paths()['include']]

        self.__parameter_values = [np.array([]) for _ in range(n_phases)]
        self.__p_global_values = np.array([])
        self.__problem_class = "MOCP"
        self.__json_file = 'mocp.json'

        self.code_export_directory = 'c_generated_code'
        """Path to where code will be exported. Default: `c_generated_code`."""

        self.simulink_opts = None
        """Options to configure Simulink S-function blocks, mainly to activate possible Inputs and Outputs."""


    @property
    def parameter_values(self):
        """:math:`p` - list of initial values for parameter vector.
        Type: `list` of `numpy.ndarray` of shape `(np_i, )`.
        - can be updated stagewise."""
        return self.__parameter_values

    @parameter_values.setter
    def parameter_values(self, parameter_values):
        if not isinstance(parameter_values, list):
            raise Exception('parameter_values must be a list of numpy.ndarrays.')
        elif len(parameter_values) != self.n_phases:
            raise Exception('parameter_values must be a list of length n_phases.')
        self.__parameter_values = parameter_values


    @property
    def p_global_values(self):
        r"""initial values for :math:`p_\text{global}` vector, see `AcadosModel.p_global` - can be updated.
        NOTE: `p_global` is shared between all phases.
        Type: `numpy.ndarray` of shape `(np_global, )`.
        """
        return self.__p_global_values

    @p_global_values.setter
    def p_global_values(self, p_global_values):
        if not isinstance(p_global_values, np.ndarray):
            raise Exception('p_global_values must be a single numpy.ndarrays.')
        self.__p_global_values = p_global_values


    @property
    def json_file(self):
        """Name of the json file where the problem description is stored."""
        return self.__json_file

    @json_file.setter
    def json_file(self, json_file):
        self.__json_file = json_file

    def set_phase(self, ocp: AcadosOcp, phase_idx: int) -> None:
        """
        Set phase of the multiphase OCP to match the given OCP.

        NOTE: model, cost, constraints and parameter_values are taken from phase OCP,
              all other fields, especially options are ignored.

        :param ocp: OCP to be set as phase
        :param phase_idx: index of the phase, must be in [0, n_phases-1]
        """
        if phase_idx >= self.n_phases:
            raise Exception(f"phase_idx {phase_idx} out of bounds, must be in [0, {self.n_phases-1}].")

        # check options
        non_default_opts = find_non_default_fields_of_obj(ocp.solver_options)
        if len(non_default_opts) > 0:
            print(f"WARNING: set_phase: Phase {phase_idx} contains non-default solver options: {non_default_opts}, which will be ignored.\n",
                   "Solver options need to be set via AcadosMultiphaseOcp.solver_options or mocp_opts instead.")

        # set phase
        self.model[phase_idx] = ocp.model
        self.cost[phase_idx] = ocp.cost
        self.constraints[phase_idx] = ocp.constraints
        self.parameter_values[phase_idx] = ocp.parameter_values

        if ocp.p_global_values.size > 0:
            print(f"WARNING: set_phase: Phase {phase_idx} contains p_global_values which will be ignored.")

        return

    def make_consistent(self) -> None:

        self.N_horizon = sum(self.N_list)
        self.solver_options.N_horizon = self.N_horizon # NOTE: to not change options when making ocp consistent

        # check options
        self.mocp_opts.make_consistent(self.solver_options, n_phases=self.n_phases)

        # check phases formulation objects are distinct
        warning = "\nNOTE: this can happen if set_phase() is called with the same ocp object for multiple phases."
        for field in ['model', 'cost', 'constraints']:
            if len(set(getattr(self, field))) != self.n_phases:
                raise Exception(f"AcadosMultiphaseOcp: make_consistent: {field} objects are not distinct.{warning}")

        # p_global check:
        p_global = self.model[0].p_global
        for i in range(self.n_phases):
            if is_empty(p_global) and not is_empty(self.model[i].p_global):
                raise Exception(f"p_global is empty for phase 0, but not for phase {i}. Should be the same for all phases.")
            if not is_empty(p_global) and not ca.is_equal(p_global, self.model[i].p_global):
                raise Exception(f"p_global is different for phase 0 and phase {i}. Should be the same for all phases.")

        # compute phase indices
        phase_idx = np.cumsum([0] + self.N_list).tolist()

        self.start_idx = phase_idx[:-1]
        self.end_idx = phase_idx[1:]

        self.cost_start_idx = phase_idx.copy()
        self.cost_start_idx[0] += 1

        # make model names unique if necessary
        model_name_list = [self.model[i].name for i in range(self.n_phases)]
        n_names = len(set(model_name_list))
        if n_names != self.n_phases:
            print(f"model names are not unique: got {model_name_list}")
            print("adding _i to model names")
            for i in range(self.n_phases):
                self.model[i].name = f"{self.model[i].name}_{i}"
            model_name_list = [self.model[i].name for i in range(self.n_phases)]
            print(f"new model names are {model_name_list}")

        # make phase OCPs consistent, warn about unused fields
        for i in range(self.n_phases):
            ocp = AcadosOcp()
            ocp.dims = self.phases_dims[i]
            ocp.model = self.model[i]
            ocp.constraints = self.constraints[i]
            ocp.cost = self.cost[i]
            ocp.parameter_values = self.parameter_values[i]
            ocp.p_global_values = self.p_global_values
            ocp.solver_options = self.solver_options

            # set phase dependent options
            ocp.solver_options.integrator_type = self.mocp_opts.integrator_type[i]
            ocp.solver_options.collocation_type = self.mocp_opts.collocation_type[i]
            ocp.solver_options.cost_discretization = self.mocp_opts.cost_discretization[i]

            if i != self.n_phases - 1:
                nondefault_fields = []
                nondefault_fields += find_non_default_fields_of_obj(ocp.cost, stage_type='terminal')
                nondefault_fields += find_non_default_fields_of_obj(ocp.constraints, stage_type='terminal')
                nondefault_fields += find_non_default_fields_of_obj(ocp.model, stage_type='terminal')
                if len(nondefault_fields) > 0:
                    print(f"Phase {i} contains non-default terminal fields: {nondefault_fields}, which will be ignored.")
            elif i != 0:
                nondefault_fields = []
                nondefault_fields += find_non_default_fields_of_obj(ocp.cost, stage_type='initial')
                nondefault_fields += find_non_default_fields_of_obj(ocp.constraints, stage_type='initial')
                nondefault_fields += find_non_default_fields_of_obj(ocp.model, stage_type='initial')
                if len(nondefault_fields) > 0:
                    print(f"Phase {i} contains non-default initial fields: {nondefault_fields}, which will be ignored.")

            print(f"Calling make_consistent for phase {i}.")
            ocp.make_consistent(is_mocp_phase=True)

            self.dummy_ocp_list.append(ocp)

        # check for transition consistency
        nx_list = [self.phases_dims[i].nx for i in range(self.n_phases)]
        for i in range(1, self.n_phases):
            if nx_list[i] != nx_list[i-1]:
                if self.phases_dims[i].nx != self.phases_dims[i-1].nx_next:
                    raise Exception(f"detected stage transition with different nx from phase {i-1} to {i}, nx_next at phase {i-1} = {self.phases_dims[i-1].nx_next} should match nx at phase {i} = {nx_list[i]}.")
                if self.N_list[i-1] != 1 or self.mocp_opts.integrator_type[i-1] != 'DISCRETE':
                    raise Exception(f"detected stage transition with different nx from phase {i-1} to {i}, which is only supported for integrator_type='DISCRETE' and N_list[i] == 1.")
        return


    def to_dict(self) -> dict:
        # Copy ocp object dictionary
        ocp_dict = dict(deepcopy(self).__dict__)
        del ocp_dict['dummy_ocp_list']

        # convert acados classes to dicts
        for key, v in ocp_dict.items():
            if isinstance(v, (AcadosOcpOptions, AcadosMultiphaseOptions)):
                ocp_dict[key]=dict(getattr(self, key).__dict__)
            if isinstance(v, list):
                for i, item in enumerate(v):
                    if isinstance(item, (AcadosModel, AcadosOcpDims, AcadosOcpConstraints, AcadosOcpCost)):
                        ocp_dict[key][i] = format_class_dict(dict(item.__dict__))

        ocp_dict = format_class_dict(ocp_dict)

        # delete keys that should not be used
        del ocp_dict['solver_options']['integrator_type']
        del ocp_dict['solver_options']['collocation_type']
        del ocp_dict['solver_options']['cost_discretization']

        return ocp_dict


    def dump_to_json(self) -> None:
        ocp_nlp_dict = self.to_dict()
        with open(self.json_file, 'w') as f:
            json.dump(ocp_nlp_dict, f, default=make_object_json_dumpable, indent=4, sort_keys=True)
        return


    def __get_template_list(self, cmake_builder=None) -> list:
        """
        returns a list of tuples in the form:
        (input_filename, output_filname)
        or
        (input_filename, output_filname, output_directory)
        """
        name = self.name
        template_list = []

        template_list.append(('main_multi.in.c', f'main_{name}.c'))
        template_list.append(('acados_multi_solver.in.h', f'acados_solver_{name}.h'))
        template_list.append(('acados_multi_solver.in.c', f'acados_solver_{name}.c'))
        # template_list.append(('acados_solver.in.pxd', f'acados_solver.pxd'))
        if cmake_builder is not None:
            template_list.append(('multi_CMakeLists.in.txt', 'CMakeLists.txt'))
        else:
            template_list.append(('multi_Makefile.in', 'Makefile'))

        if self.phases_dims[0].np_global > 0:
            template_list.append(('p_global_precompute_fun.in.h', f'{self.name}_p_global_precompute_fun.h'))

        # Simulink
        if self.simulink_opts is not None:
            raise NotImplementedError('Simulink not yet supported for multiphase OCPs.')

        return template_list


    def render_templates(self, cmake_builder=None):

        # model templates
        for i, dummy_ocp in enumerate(self.dummy_ocp_list):
            # this is the only option that can vary and influence external functions to be generated
            dummy_ocp.solver_options.integrator_type = self.mocp_opts.integrator_type[i]

            template_list = dummy_ocp._get_external_function_header_templates()
            # dump dummy_ocp
            dummy_ocp.json_file = 'tmp_ocp.json'
            dummy_ocp.dump_to_json()
            tmp_json_path = os.path.abspath(dummy_ocp.json_file)

            # render templates
            for tup in template_list:
                output_dir = self.code_export_directory if len(tup) <= 2 else tup[2]
                render_template(tup[0], tup[1], output_dir, tmp_json_path)

        print("rendered model templates successfully")

        # check json file
        json_path = os.path.abspath(self.json_file)
        if not os.path.exists(json_path):
            raise Exception(f'Path "{json_path}" not found!')

        # solver templates
        template_list = self.__get_template_list(cmake_builder=cmake_builder)

        # Render templates
        for tup in template_list:
            output_dir = self.code_export_directory if len(tup) <= 2 else tup[2]
            render_template(tup[0], tup[1], output_dir, json_path)

        # # Custom templates
        # acados_template_path = os.path.dirname(os.path.abspath(__file__))
        # custom_template_glob = os.path.join(acados_template_path, 'custom_update_templates', '*')
        # for tup in ocp.solver_options.custom_templates:
        #     render_template(tup[0], tup[1], ocp.code_export_directory, json_path, template_glob=custom_template_glob)
        print("\nmocp_render_templates: rendered solver templates successfully!\n")

        return


    def generate_external_functions(self) -> GenerateContext:

        # options for code generation
        code_gen_opts = dict()
        code_gen_opts['generate_hess'] = self.solver_options.hessian_approx == 'EXACT'
        code_gen_opts['with_solution_sens_wrt_params'] = self.solver_options.with_solution_sens_wrt_params
        code_gen_opts['with_value_sens_wrt_params'] = self.solver_options.with_value_sens_wrt_params
        code_gen_opts['code_export_directory'] = self.code_export_directory

        context = GenerateContext(self.model[0].p_global, self.name, code_gen_opts)

        for i in range(self.n_phases):
            ignore_initial = True if i != 0 else False
            ignore_terminal = True if i != self.n_phases-1 else False
            # this is the only option that can vary and influence external functions to be generated
            self.dummy_ocp_list[i].solver_options.integrator_type = self.mocp_opts.integrator_type[i]
            context = self.dummy_ocp_list[i]._setup_code_generation_context(context, ignore_initial, ignore_terminal)
            self.dummy_ocp_list[i].code_export_directory = self.code_export_directory

        context.finalize()
        self.__external_function_files_model = context.get_external_function_file_list(ocp_specific=False)
        self.__external_function_files_ocp = context.get_external_function_file_list(ocp_specific=True)
        for i in range(self.n_phases):
            self.phases_dims[i].n_global_data = context.get_n_global_data()

        return context



#
# Copyright (c) The acados authors.
#
# This file is part of acados.
#
# The 2-Clause BSD License
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
# this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.;
#

from casadi import *
import numpy as np
from ..utils import casadi_length, idx_perm_to_ipiv
from .determine_input_nonlinearity_function import determine_input_nonlinearity_function
from .check_reformulation import check_reformulation


def determine_trivial_gnsf_transcription(acados_ocp, print_info):
    ## Description
    # this function takes a model of an implicit ODE/ index-1 DAE and sets up
    # an equivalent model in the GNSF structure, with empty linear output
    # system and trivial model matrices, i.e. A, B, E, c are zeros, and C is
    # eye. - no structure is exploited

    model = acados_ocp.model
    # initial print
    print("*****************************************************************")
    print(" ")
    print(f"******      Restructuring {model.name} model    ***********")
    print(" ")
    print("*****************************************************************")

    # load model
    f_impl_expr = model.f_impl_expr

    model_name_prefix = model.name

    # x
    x = model.x
    nx = acados_ocp.dims.nx
    # check type
    if isinstance(x[0], SX):
        isSX = True
    else:
        print("GNSF detection only works for SX CasADi type!!!")
        import pdb

        pdb.set_trace()
    # xdot
    xdot = model.xdot
    # u
    nu = acados_ocp.dims.nu
    if nu == 0:
        u = SX.sym("u", 0, 0)
    else:
        u = model.u

    nz = acados_ocp.dims.nz
    if nz == 0:
        z = SX.sym("z", 0, 0)
    else:
        z = model.z

    p = model.p
    nparam = acados_ocp.dims.np

    # avoid SX of size 0x1
    if casadi_length(u) == 0:
        u = SX.sym("u", 0, 0)
        nu = 0
    ## initialize gnsf struct
    # dimensions
    gnsf = {"nx": nx, "nu": nu, "nz": nz, "np": nparam}
    gnsf["nx1"] = nx
    gnsf["nx2"] = 0
    gnsf["nz1"] = nz
    gnsf["nz2"] = 0
    gnsf["nuhat"] = nu
    gnsf["ny"] = 2 * nx + nz

    gnsf["phi_expr"] = f_impl_expr
    gnsf["A"] = np.zeros((nx + nz, nx))
    gnsf["B"] = np.zeros((nx + nz, nu))
    gnsf["E"] = np.zeros((nx + nz, nx + nz))
    gnsf["c"] = np.zeros((nx + nz, 1))
    gnsf["C"] = np.eye(nx + nz)
    gnsf["name"] = model_name_prefix

    gnsf["x"] = x
    gnsf["xdot"] = xdot
    gnsf["z"] = z
    gnsf["u"] = u
    gnsf["p"] = p

    gnsf = determine_input_nonlinearity_function(gnsf)

    gnsf["A_LO"] = []
    gnsf["E_LO"] = []
    gnsf["B_LO"] = []
    gnsf["c_LO"] = []
    gnsf["f_lo_expr"] = []

    # permutation
    gnsf["idx_perm_x"] = range(nx)  # matlab-style)
    gnsf["ipiv_x"] = idx_perm_to_ipiv(gnsf["idx_perm_x"])  # blasfeo-style
    gnsf["idx_perm_z"] = range(nz)
    gnsf["ipiv_z"] = idx_perm_to_ipiv(gnsf["idx_perm_z"])
    gnsf["idx_perm_f"] = range((nx + nz))
    gnsf["ipiv_f"] = idx_perm_to_ipiv(gnsf["idx_perm_f"])

    gnsf["nontrivial_f_LO"] = 0

    check_reformulation(model, gnsf, print_info)
    if print_info:
        print(f"Success: Set up equivalent GNSF model with trivial matrices")
        print(" ")
    if print_info:
        print(
            "-----------------------------------------------------------------------------------"
        )
        print(" ")
        print(
            "reduced input ny    of phi from  ",
            str(2 * nx + nz),
            "   to  ",
            str(gnsf["ny"]),
        )
        print(
            "reduced input nuhat of phi from  ", str(nu), "   to  ", str(gnsf["nuhat"])
        )
        print(" ")
        print(
            "-----------------------------------------------------------------------------------"
        )
    return gnsf



#
#     This file is part of rockit.
#
#     rockit -- Rapid Optimal Control Kit
#     Copyright (C) 2019 MECO, KU Leuven. All rights reserved.
#
#     Rockit is free software; you can redistribute it and/or
#     modify it under the terms of the GNU Lesser General Public
#     License as published by the Free Software Foundation; either
#     version 3 of the License, or (at your option) any later version.
#
#     Rockit is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#     Lesser General Public License for more details.
#
#     You should have received a copy of the GNU Lesser General Public
#     License along with CasADi; if not, write to the Free Software
#     Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
#
#
from casadi import vertcat, vcat, which_depends, MX, substitute, integrator, Function, depends_on
from .stage import Stage, transcribed
from .placeholders import TranscribedPlaceholders
from .casadi_helpers import vvcat, rockit_pickle_context, rockit_unpickle_context
from .external.manager import external_method
from .direct_method import DirectMethod
class Ocp(Stage):
    def __init__(self,  t0=0, T=1, scale=1, **kwargs):
        """Create an Optimal Control Problem environment

        Parameters
        ----------
        t0 : float or :obj:`~rockit.freetime.FreeTime`, optional
            Starting time of the optimal control horizon
            Default: 0
        T : float or :obj:`~rockit.freetime.FreeTime`, optional
            Total horizon of the optimal control horizon
            Default: 1
        scale: float, optional
               Typical time scale

        Examples
        --------

        >>> ocp = Ocp()
        """
        Stage.__init__(self,  t0=t0, T=T, scale=scale, **kwargs)
        self._master = self
        # Flag to make solve() faster when solving a second time
        # (e.g. with different parameter values)
        self._var_is_transcribed = False
        self._transcribed_placeholders = TranscribedPlaceholders()

    @transcribed
    def jacobian(self, with_label=False):
        return self._method.jacobian(with_label=with_label)

    @transcribed
    def hessian(self, with_label=False):
        return self._method.hessian(with_label=with_label)

    @transcribed
    def spy_jacobian(self):
        self._method.spy_jacobian()

    @transcribed
    def spy_hessian(self):
        self._method.spy_hessian()

    @transcribed
    def spy(self):
        import matplotlib.pylab as plt
        plt.subplots(1, 2, figsize=(10, 4))
        plt.subplot(1, 2, 1)
        self.spy_jacobian()
        plt.subplot(1, 2, 2)
        self.spy_hessian()

    @property
    def _transcribed(self):
        if self._is_original:
            if self._is_transcribed:
                return self._augmented
            else:

                import copy
                augmented = copy.deepcopy(self)
                augmented._master = augmented
                augmented._transcribed_placeholders = self._transcribed_placeholders
                augmented._var_original = self
                self._var_augmented = augmented
                augmented._placeholders = self._placeholders
                
                return self._augmented._transcribed
        else:
            self._transcribe()
            return self
        
    def transcribe(self,**kwargs):
        self._untranscribe()
        self._transcribe(**kwargs)

    def _transcribe(self,**kwargs):
        if not self.is_transcribed:
            self._transcribed_placeholders.clear()
            self._transcribe_recurse(phase=0,**kwargs)
            self._placeholders_transcribe_recurse(1,self._transcribed_placeholders)
            self._transcribe_recurse(phase=1,**kwargs)
            self._original._set_transcribed(True)

            self._transcribe_recurse(phase=2,placeholders=self.placeholders_transcribed,**kwargs)
    
    def _untranscribe(self,**kwargs):
        if self.is_transcribed:
            self._transcribed_placeholders.clear()
            self._untranscribe_recurse(phase=0)
            self._placeholders_untranscribe_recurse(1)
            self._untranscribe_recurse(phase=1)
            self._original._set_transcribed(False)

            self._untranscribe_recurse(phase=2)

    @property
    @transcribed
    def placeholders_transcribed(self):
        """
        May also be called after solving (issue #91)
        """
        if self._transcribed_placeholders.is_dirty:
            self._placeholders_transcribe_recurse(2, self._transcribed_placeholders)
            self._transcribed_placeholders.is_dirty = False
        return self._transcribed_placeholders

    @property
    def non_converged_solution(self):
        return self._method.non_converged_solution(self)

    @transcribed
    def solve(self):
        return self._method.solve(self)
 
    @transcribed
    def solve_limited(self):
        return self._method.solve_limited(self)

    def callback(self, fun):
        self._set_transcribed(False)
        return self._method.callback(self, fun)

    @property
    def debug(self):
        self._method.debug

    def solver(self, solver, solver_options={}):
        self._method.solver(solver, solver_options)

    def show_infeasibilities(self, *args):
        self._method.show_infeasibilities(*args)

    def debugme(self,e):
        print(e,hash(e),e.__hash__())
        return e

    @property
    @transcribed
    def gist(self):
        """Obtain an expression packing all information needed to obtain value/sample

        The composition of this array may vary between rockit versions

        Returns
        -------
        :obj:`~casadi.MX` column vector

        """
        return self._method.gist

    @transcribed
    def to_function(self, name, args, results, *margs):
        return self._method.to_function(self, name, args, results, *margs)

    def is_sys_time_varying(self):
        # For checks
        self._ode()
        ode = vvcat([self._state_der[k] for k in self.states])
        alg = vvcat(self._alg)
        rhs = vertcat(ode,alg)
        return depends_on(rhs, self.t)

    def is_parameter_appearing_in_sys(self):
        # For checks
        self._ode()
        ode = vvcat([self._state_der[k] for k in self.states])
        alg = vvcat(self._alg)
        rhs = vertcat(ode,alg)
        pall = self.parameters['']+self.parameters['control']
        dep = [depends_on(rhs,p) for p in pall]
        return dep

    def sys_dae(self):
        # For checks
        self._ode()
        ode = vvcat([self._state_der[k] for k in self.states])
        alg = vvcat(self._alg)

        dae = {}
        dae["x"] = self.x
        dae["z"] = self.z

        t0 = MX.sym("t0")
        dt = MX.sym("dt")
        tau = MX.sym("tau")
        dae["t"] = self.t

        pall = self.parameters['']+self.parameters['control']

        rhs = vertcat(ode,alg)
        dep = [depends_on(rhs,p) for p in pall]

        p = vvcat([p for p,dep in zip(pall,dep) if dep])

        dae["ode"] = ode
        dae["alg"] = alg
        dae["p"] = p
        dae["u"] = self.u
        return dae

    def view_api(self, name):
        method = self._method
        self.method(external_method(name,method=method))
        self._transcribed
        return self._method

    def sys_simulator(self, intg='rk', intg_options=None):
        if intg_options is None:
            intg_options = {}
        intg_options = dict(intg_options)
        # For checks
        self._ode()
        ode = vvcat([self._state_der[k] for k in self.states])
        alg = vvcat(self._alg)

        dae = {}
        dae["x"] = self.x
        dae["z"] = self.z

        t0 = MX.sym("t0")
        dt = MX.sym("dt")
        tau = MX.sym("tau")
        dae["t"] = tau

        pall = self.parameters['']+self.parameters['control']
        
        #dep = which_depends(vertcat(ode,alg),vvcat(pall),1,True)


        rhs = vertcat(ode,alg)
        dep = [depends_on(rhs,p) for p in pall]

        p = vvcat([p for p,dep in zip(pall,dep) if dep])

        [ode,alg] = substitute([ode,alg],[self.t],[t0+tau*dt])
        dae["ode"] = dt*ode
        dae["alg"] = alg
        dae["p"] = vertcat(self.u, t0, dt, p)

        try:
            intg = integrator('intg',intg,dae,0,1,intg_options)
        except:
            intg_options["t0"] = 0
            intg_options["tf"] = 1
            intg = integrator('intg',intg,dae,intg_options)

        z_initial_guess = MX.sym("z",self.z.sparsity()) if self.nz>0 else MX(0,1)
        
        intg_out = intg(x0=self.x, p=dae["p"], z0=z_initial_guess)
        simulator = Function('simulator',
            [self.x, self.u, p, t0, dt, z_initial_guess],
            [intg_out["xf"], intg_out["zf"]],
            ["x","u","p","t0","dt","z_initial_guess"],
            ["xf","zf"])
        return simulator

    def save(self,name):
        self._untranscribe()
        import pickle
        with rockit_pickle_context():
            pickle.dump(self,open(name,"wb"))

    @staticmethod
    def load(name):
        import pickle
        with rockit_unpickle_context():
            return pickle.load(open(name,"rb"))


#
# Copyright (c) The acados authors.
#
# This file is part of acados.
#
# The 2-Clause BSD License
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
# this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.;
#

cimport acados_sim_solver_common

cdef extern from "acados_sim_solver_{{ model.name }}.h":
    ctypedef struct sim_solver_capsule "{{ model.name }}_sim_solver_capsule":
        pass

    sim_solver_capsule * acados_sim_solver_create_capsule "{{ model.name }}_acados_sim_solver_create_capsule"()
    int acados_sim_solver_free_capsule "{{ model.name }}_acados_sim_solver_free_capsule"(sim_solver_capsule *capsule)

    int acados_sim_create "{{ model.name }}_acados_sim_create"(sim_solver_capsule * capsule)
    int acados_sim_solve "{{ model.name }}_acados_sim_solve"(sim_solver_capsule * capsule)
    int acados_sim_free "{{ model.name }}_acados_sim_free"(sim_solver_capsule * capsule)
    int acados_sim_update_params "{{ model.name }}_acados_sim_update_params"(sim_solver_capsule * capsule, double *value, int np_)
    # int acados_sim_update_params_sparse "{{ model.name }}_acados_sim_update_params_sparse"(sim_solver_capsule * capsule, int stage, int *idx, double *p, int n_update)

    acados_sim_solver_common.sim_in *acados_get_sim_in "{{ model.name }}_acados_get_sim_in"(sim_solver_capsule * capsule)
    acados_sim_solver_common.sim_out *acados_get_sim_out "{{ model.name }}_acados_get_sim_out"(sim_solver_capsule * capsule)
    acados_sim_solver_common.sim_solver *acados_get_sim_solver "{{ model.name }}_acados_get_sim_solver"(sim_solver_capsule * capsule)
    acados_sim_solver_common.sim_config *acados_get_sim_config "{{ model.name }}_acados_get_sim_config"(sim_solver_capsule * capsule)
    acados_sim_solver_common.sim_opts *acados_get_sim_opts "{{ model.name }}_acados_get_sim_opts"(sim_solver_capsule * capsule)
    void *acados_get_sim_dims "{{ model.name }}_acados_get_sim_dims"(sim_solver_capsule * capsule)



#
#     This file is part of rockit.
#
#     rockit -- Rapid Optimal Control Kit
#     Copyright (C) 2019 MECO, KU Leuven. All rights reserved.
#
#     Rockit is free software; you can redistribute it and/or
#     modify it under the terms of the GNU Lesser General Public
#     License as published by the Free Software Foundation; either
#     version 3 of the License, or (at your option) any later version.
#
#     Rockit is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#     Lesser General Public License for more details.
#
#     You should have received a copy of the GNU Lesser General Public
#     License along with CasADi; if not, write to the Free Software
#     Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
#
#

from .sampling_method import SamplingMethod
from casadi import sumsqr, horzcat, vertcat, linspace, substitute, MX, evalf,\
                   vcat, collocation_points, collocation_interpolators, hcat,\
                   repmat, DM, sum2, mtimes, vvcat, depends_on, Function
from .casadi_helpers import get_ranges_dict, HashOrderedDict, HashDict, is_numeric
import casadi as ca
from itertools import repeat
try:
    from casadi import collocation_coeff
except:
    def collocation_coeff(tau):
        [C, D] = collocation_interpolators(tau)
        d = len(tau)
        tau = [0]+tau
        F = [None]*(d+1)
        # Construct polynomial basis
        for j in range(d+1):
            # Construct Lagrange polynomials to get the polynomial basis at the collocation point
            p = np.poly1d([1])
            for r in range(d+1):
                if r != j:
                    p *= np.poly1d([1, -tau[r]]) / (tau[j]-tau[r])
            
            # Evaluate the integral of the polynomial to get the coefficients of the quadrature function
            pint = np.polyint(p)
            F[j] = pint(1.0)
    
        return (hcat(C[1:]), vcat(D), hcat(F[1:]))

import numpy as np

class DirectCollocation(SamplingMethod):
    def __init__(self, degree=4, scheme='radau', **kwargs):
        SamplingMethod.__init__(self, **kwargs)
        self.degree = degree
        self.tau = collocation_points(degree, scheme)
        [self.C, self.D, self.B] = collocation_coeff(self.tau)
        self.clean()

    def clean(self):
        SamplingMethod.clean(self)
        self.Zc = []  # List that will hold algebraic decision variables list(N,list(M,nz x degree))
        self.Xc = []  # List that will hold helper collocation states list(N,list(M, degree+1))
        self.X_intg = []
        self.Xc_pure = []
        self.Zc0 = []
        self.Xc_vars = []
        self.Xc_vars0 = []
        self.Zc_vars_base = []
        self.Zc_vars_rest = []


    def add_variables(self, stage, opti):
        scale_x = stage._scale_x
        scale_z = stage._scale_z
        scale_u = stage._scale_u

        # We are creating variables in a special order such that the resulting constraint Jacobian
        # is block-sparse
        x = opti.variable(stage.nx, scale=scale_x)
        self.X.append(x)
        self.Q.append(DM.zeros(stage.nxq))
        self.add_variables_V(stage, opti)
        z = opti.variable(stage.nz, scale=scale_z)
        self.Zc_vars_base.append(z)

        for k in range(self.N):
            self.U.append(opti.variable(stage.nu, scale=scale_u) if stage.nu>0 else MX(0,1))
            xr = []
            zr = []
            Xc = []
            Zc = []
            for i in range(self.M):
                xc = opti.variable(stage.nx, self.degree, scale=repmat(scale_x, 1, self.degree))
                xr.append(xc)
                zc = opti.variable(stage.nz, self.degree-1, scale=repmat(scale_z, 1, self.degree-1))
                x0 = x if i==0 else opti.variable(stage.nx, scale=scale_x)
                self.X_intg.append(x0)
                self.Xc_pure.append(xc)
                Xc.append(horzcat(x0, xc))
                self.Xc_vars.append(xc if i==0 else horzcat(x0, xc))
                self.Xc_vars0.append(repmat(x, 1, self.degree if i==0 else self.degree+1))
                z0 = z if i==0 else opti.variable(stage.nz, scale=scale_z)
                zr.append(horzcat(z0, zc))
                Zc.append(horzcat(z0,zc))
                self.Zc_vars_rest.append(zc if i==0 else horzcat(z0, zc))
                self.Zc0.append(repmat(z, 1, self.degree-1 if i==0 else self.degree))

            self.xr.append(xr)
            self.zr.append(zr)
            self.Xc.append(Xc)
            self.Zc.append(Zc)
            x = opti.variable(stage.nx, scale=scale_x)
            self.X.append(x)
            self.Q.append(None)
            z = opti.variable(stage.nz, scale=scale_z)
            self.Zc_vars_base.append(z)
            self.add_variables_V_control(stage, opti, k)

        self.Xc_vars = hcat(self.Xc_vars)
        self.Xc_vars0 = hcat(self.Xc_vars0)
        self.Zc0 = hcat(self.Zc0)
        self.Zc_vars_base = hcat(self.Zc_vars_base)
        self.Zc_vars_rest = hcat(self.Zc_vars_rest)

        self.add_variables_V_control_finalize(stage, opti)

    def add_constraints(self, stage, opti):
        self.add_constraints_before(stage, opti)
        scale_x = stage._scale_x
        scale_der_x = stage._scale_der_x
        scale_z = stage._scale_z
        self.poly_coeff_q = None

        # Obtain the discretised system
        f = stage._ode()

        ps = []
        tau_root = [0] + self.tau
        # Construct polynomial basis
        for j in range(self.degree + 1):
            # Construct Lagrange polynomials to get the polynomial basis at the collocation point
            p = np.poly1d([1])
            for r in range(self.degree + 1):
                if r != j:
                    p *= np.poly1d([1, -tau_root[r]]) / (tau_root[j] - tau_root[r])
            ps.append(hcat(p.coef[::-1]))
        poly = vcat(ps)
        ps_z = []
        # Construct polynomial basis for z cfr "2.5 Continuous Output for Optimal Control" from Rien's thesis
        for j in range(1, self.degree + 1):
            # Construct Lagrange polynomials to get the polynomial basis at the collocation point
            p_z = np.poly1d([1])
            for r in range(1, self.degree + 1):
                if r != j:
                    p_z *= np.poly1d([1, -tau_root[r]]) / (tau_root[j] - tau_root[r])
            ps_z.append(hcat(p_z.coef[::-1]))
        poly_z = vcat(ps_z)
        self.q = 0

        # Make time-grid for roots
        for k in range(self.N):
            dt = (self.control_grid[k + 1] - self.control_grid[k])/self.M
            tr = []
            for i in range(self.M):
                tr.append([self.integrator_grid[k][i]+dt*self.tau[j] for j in range(self.degree)])        
            self.tr.append(tr)

        # Handle Bspline signals
        subgrid = []
        for i in range(self.M):
            subgrid+=list((i+np.array(self.tau))/self.M)

        v_sampled_store = []
        for e in self.signals.values():
            v_sampled = ca.horzsplit(e.sample(subgrid=subgrid,include_edges=False))
            v_sampled_store.append(v_sampled)
        signals_sampled = []
        for i in range(len(subgrid*self.N)):
            signals_sampled.append(ca.vertcat(*[e[i] for e in v_sampled_store]))

        dts = []
        # Fill in Z variables up-front, since they might be needed in constraints with ocp.next
        for k in range(self.N):
            dt = (self.control_grid[k + 1] - self.control_grid[k])/self.M
            dts.append(dt)
            S = 1/repmat(hcat([dt**i for i in range(self.degree + 1)]), self.degree + 1, 1)
            S_z = 1/repmat(hcat([dt**i for i in range(self.degree)]), self.degree, 1)
            self.Z.append(mtimes(self.Zc[k][0],poly_z[:,0]))
            for i in range(self.M):
                self.xk.append(self.Xc[k][i][:,0])
                self.zk.append(mtimes(self.Zc[k][i],poly_z[:,0]))
                self.poly_coeff.append(mtimes(self.Xc[k][i],poly*S))
                self.poly_coeff_z.append(mtimes(self.Zc[k][i],poly_z*S_z))

        self.Z.append(mtimes(self.Zc[-1][-1],sum2(poly_z)))

        self.xqk.append(DM.zeros(stage.nxq))
        count_f_eval = 0
        for k in range(self.N):
            dt = dts[k]
            p = self.get_p_sys(stage,k,include_signals=False)
            for i in range(self.M):
                for j in range(self.degree):
                    Pidot_j = mtimes(self.Xc[k][i],self.C[:,j])/ dt
                    p_total = vertcat(p, signals_sampled[count_f_eval])
                    res = f(x=self.Xc[k][i][:, j+1], u=self.U[k], z=self.Zc[k][i][:,j], p=p_total, t=self.tr[k][i][j])
                    count_f_eval += 1
                    # Collocation constraints
                    opti.subject_to(Pidot_j == res["ode"], scale=scale_der_x)
                    self.q = self.q + res["quad"]*dt*self.B[j]
                    if stage.nz:
                        opti.subject_to(0 == res["alg"], scale = scale_z)
                    for c, meta, args in stage._constraints["integrator_roots"]:
                        opti.subject_to(self.eval_at_integrator_root(stage, c, k, i, j), scale=args["scale"], meta=meta)

                # Continuity constraints
                x_next = self.X[k + 1] if i==self.M-1 else self.Xc[k][i+1][:,0]
                opti.subject_to(mtimes(self.Xc[k][i],self.D) == x_next, scale=scale_x)

                for c, meta, args in stage._constraints["integrator"]:
                    if k==0 and i==0 and not args["include_first"]: continue
                    opti.subject_to(self.eval_at_integrator(stage, c, k, i), scale=args["scale"], meta=meta)
                for c, meta, _ in stage._constraints["inf"]:
                    self.add_inf_constraints(stage, opti, c, k, i, meta)
                self.xqk.append(self.q)
            for c, meta, args in stage._constraints["control"]:  # for each constraint expression
                if k==0 and not args["include_first"]: continue
                # Add it to the optimizer, but first make x,u concrete.
                try:
                    opti.subject_to(self.eval_at_control(stage, c, k), scale=args["scale"], meta=meta)
                except IndexError:
                    pass # Can be caused by ocp.offset -> drop constraint
            self.Q[k+1] = self.q
        for c, meta, args in stage._constraints["control"]:  # for each constraint expression
            if not args["include_last"]: continue
            # Add it to the optimizer, but first make x,u concrete.
            try:
                opti.subject_to(self.eval_at_control(stage, c, -1), scale=args["scale"], meta=meta)
            except IndexError:
                pass # Can be caused by ocp.offset -> drop constraint

        for c, meta, args in stage._constraints["integrator"]:  # for each constraint expression
            if not args["include_last"]: continue
            # Add it to the optimizer, but first make x,u concrete.
            opti.subject_to(self.eval_at_control(stage, c, -1), scale=args["scale"], meta=meta)

    def set_initial(self, stage, master, initial):
        opti = master.opti if hasattr(master, 'opti') else master
        opti.cache_advanced()
        initial = HashOrderedDict(initial)
        algs = get_ranges_dict(stage.algebraics)
        initial_alg = HashDict()
        for a, v in list(initial.items()):
            if a in algs:
                initial_alg[a] = v
                del initial[a]
        for var, expr in initial.items():
            if ca.is_equal(var, stage.T):
                var = self.T
            if ca.is_equal(var, stage.t0):
                var = self.t0
            is_states = depends_on(var, stage.x)
            opti_initial = opti.initial()
            if is_numeric(expr):
                value = ca.evalf(expr)
                # Row vector if vector
                if value.is_column() and var.is_scalar(): value = value.T
                if is_states:
                    if var.numel()*(self.N)==value.numel() or var.numel()*(self.N+1)==value.numel():
                        value_integrator = kron(DM.ones(1,self.M),value[:,:self.N])
                        if var.numel()*(self.N+1)==value.numel(): value_integrator = horzcat(value_integrator.value[:,-1])
                        value_integrator_root = kron(DM.ones(1,self.M*self.degree),value[:,:self.N])
                    else:
                        value_integrator = repmat(value,1,self.N*self.M+1)
                        value_integrator_root = repmat(value,1,self.N*self.M*self.degree)
            else:
                if is_states:
                    NUM = vertcat(DM(range(10)).T, DM(range(10,20)).T)
                    expr_integrator = ca.hcat([self.eval_at_integrator(stage, expr, k, i) for k in list(range(self.N)) for i in range(self.M)]+[self.eval_at_control(stage, expr, -1)]) # HOT line
                    expr_integrator_root = ca.hcat([self.eval_at_integrator_root(stage, expr, k, i, j) for k in list(range(self.N)) for i in range(self.M) for j in range(self.degree) ]) # HOT line
                    value_integrator = DM(opti.debug.value(expr_integrator, opti_initial))
                    value_integrator_root = DM(opti.debug.value(expr_integrator_root, opti_initial))
                else:
                    expr = ca.hcat([self.eval_at_control(stage, expr, k) for k in list(range(self.N))+[-1]]) # HOT line
                    value = DM(opti.debug.value(expr, opti_initial))

            if is_states:
                target_integrator = ca.hcat([self.eval_at_integrator(stage, var, k, i) for k in list(range(self.N)) for i in range(self.M)]+[self.eval_at_control(stage, var, -1)])
                opti.set_initial(target_integrator, value_integrator, cache_advanced=True)
                target_integrator_root = ca.hcat([self.eval_at_integrator_root(stage, var, k, i, j) for k in list(range(self.N)) for i in range(self.M) for j in range(self.degree)]) # HOT line
                opti.set_initial(target_integrator_root, value_integrator_root, cache_advanced=True)
            else:
                # Row vector if vector
                if value.is_column() and var.is_scalar(): value = value.T
                for k in list(range(self.N))+[-1]:
                    target = self.eval_at_control(stage, var, k)
                    value_k = value
                    if target.numel()*(self.N)==value.numel() or target.numel()*(self.N+1)==value.numel():
                        value_k = value[:,k]
                    try:
                        #print(target,value_k)
                        opti.set_initial(target, value_k, cache_advanced=True)
                    except Exception as e:
                        # E.g for single shooting, set_initial of a state, for k>0
                        # Error message is usually "... arbitrary expression ..." but can also be
                        # "... You cannot set an initial value for a parameter ..."
                        # if the dynamics contains a parameter
                        if "arbitrary expression" in str(e) or (not target.is_valid_input() and "initial value for a parameter" in str(e)):
                            pass
                        else:
                            # Other type of error: 
                            raise e
        for var, expr in list(initial_alg.items()):
            opti_initial = opti.initial()
            for k in range(self.N):
                for i, e in enumerate(self.Zc[k]):
                    e_shape = e[algs[var],:].shape
                    value = DM(opti.debug.value(hcat([self.eval_at_integrator_root(stage, expr, k, i, j) for j in range(e_shape[1])]), opti_initial))                    
                    opti.set_initial(e[algs[var],:], value)

    def to_function(self, stage, name, args, results, *margs):
        args = list(args)
        add_zc = 0
        for i,e in enumerate(args):
            if isinstance(e,str) and e=="z":
                args[i] = self.Zc_vars_base
                add_zc = True
        all_args = vvcat(args)
        _,states = stage.sample(stage.x,grid='control')

        name_in = None
        if len(margs)>0 and isinstance(margs[0], list) and np.all([isinstance(e,str) for e in margs[0]]):
            name_in = list(margs[0])

        add_xc = depends_on(all_args, states) and not depends_on(all_args, self.Xc_vars)
        add_zc = add_zc and not depends_on(all_args, self.Zc_vars_rest)
        inner_args = list(args)
        if add_xc:
            inner_args += [self.Xc_vars]
            if name_in: name_in += ["Xc_vars"]
        if add_zc:
            inner_args += [self.Zc_vars_rest]
            if name_in: name_in += ["Zc_vars_rest"]

        inner_margs = list(margs)
        if name_in:
            inner_margs[0] = name_in

        f = SamplingMethod.to_function(self, stage, name, inner_args, results, *inner_margs)
        f_args = f.mx_in()[:len(args)]
        call_args = list(f_args)
        if add_xc:
            call_args+=[self.Xc_vars0]
        if add_zc:
            call_args+=[self.Zc0]

        return Function(name, f_args, f.call(call_args,True,False), *margs)



#first, find a rockit way  to initialize xc and zc using to_function
#parameter?
#parametric set_initial?


import unittest

from pylab import *

from rockit import Ocp, DirectMethod, MultipleShooting, DirectCollocation, SingleShooting, SplineMethod
from problems import integrator_control_problem, bang_bang_chain_problem
import numpy as np
import casadi as ca

class IntegrTests(unittest.TestCase):

    def test_all(self):
        T = 7
        N = 10

        for intg in ['rk','cvodes']:
            ocp = Ocp(T=T)
            x_c = ocp.state()
            x_d = ocp.state()
            u = ocp.control()
            ocp.set_der(x_c, 0.3*x_d+0.8*x_c+0.7*u)
            ocp.set_next(x_d, 0.2*x_c+0.1*x_d+0.4*u)

            ocp.subject_to(ocp.at_t0(x_c)==0.7)
            ocp.subject_to(ocp.at_t0(x_d)==0.8)
            ocp.subject_to(u==ocp.t,include_last=False)

            ocp.solver('ipopt')
            ocp.method(MultipleShooting(N=N,intg=intg))
            sol = ocp.solve()

            x_c_sol = sol.sample(x_c,grid='control')[1]
            x_d_sol = sol.sample(x_d,grid='control')[1]
            u_sol = sol.sample(u,grid='control')[1]
            
            # Check correctness of discrete dynamics
            self.assertAlmostEqual(x_d_sol[0], 0.8)
            for k in range(N):
                self.assertAlmostEqual(x_d_sol[k+1], 0.2*x_c_sol[k]+0.1*x_d_sol[k]+0.4*u_sol[k])
            # Check correctness of continuous dynamics
            self.assertAlmostEqual(x_c_sol[0], 0.7)
            options = {}
            if intg == 'rk':
                options = {"number_of_finite_elements": 1}
            intg = ca.integrator('intg',intg,{"x": x_c, "p": x_d, "u": u, "ode": 0.3*x_d+0.8*x_c+0.7*u},0,T/N,options)
            for k in range(N):
                self.assertAlmostEqual(intg(x0=x_c_sol[k],p=x_d_sol[k],u=u_sol[k])["xf"], x_c_sol[k+1])

        
if __name__ == '__main__':
    unittest.main()



#
# Copyright (c) The acados authors.
#
# This file is part of acados.
#
# The 2-Clause BSD License
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
# this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.;
#

from .acados_model import AcadosModel
from .acados_dims import AcadosOcpDims, AcadosSimDims

from .acados_ocp import AcadosOcp

from .acados_ocp_cost import AcadosOcpCost
from .acados_ocp_constraints import AcadosOcpConstraints
from .acados_ocp_options import AcadosOcpOptions
from .acados_ocp_batch_solver import AcadosOcpBatchSolver
from .acados_ocp_iterate import AcadosOcpIterate, AcadosOcpIterates, AcadosOcpFlattenedIterate

from .acados_sim import AcadosSim, AcadosSimOptions
from .acados_multiphase_ocp import AcadosMultiphaseOcp

from .acados_ocp_solver import AcadosOcpSolver
from .acados_sim_solver import AcadosSimSolver
from .acados_sim_batch_solver import AcadosSimBatchSolver
from .utils import print_casadi_expression, get_acados_path, get_python_interface_path, \
    get_tera_exec_path, get_tera, check_casadi_version, acados_dae_model_json_dump, \
    casadi_length, make_object_json_dumpable, J_to_idx, get_default_simulink_opts, \
    is_empty, get_simulink_default_opts, ACADOS_INFTY

from .builders import ocp_get_default_cmake_builder, sim_get_default_cmake_builder

from .plot_utils import latexify_plot

from .penalty_utils import symmetric_huber_penalty, one_sided_huber_penalty, huber_loss

from .mpc_utils import create_model_with_cost_state

from .zoro_description import ZoroDescription






from os import stat
import casadi as cs
import os

class CodeVariable:
    def __init__(self, name, dtype,
            comment=""):
        self.name = name
        self.dtype = dtype
        self.comment = comment

class LocalVariable(CodeVariable):
    pass

class MemoryField(CodeVariable):
    def __init__(self, name, dtype,
            comment=""):
        CodeVariable.__init__(self, name, dtype, comment=comment)

class BaseCodeGenerator:
    def __init__(self,
            ocp,
            name=None,
            use_codegen=False,
            casadi_fun_name=None,
            build_dir_abs=None,
            prefix=None):
        self.ocp = ocp
        self.name = name
        self.use_codegen = use_codegen
        self.casadi_fun_name = casadi_fun_name
        self.build_dir_abs = build_dir_abs
        self.prefix = prefix

    @staticmethod
    def escape(e):
      return e.replace('\\','/')

class CasadiWrapperCodeGenerator(BaseCodeGenerator):
    def __init__(self,*args,**kwargs):
        BaseCodeGenerator.__init__(self,*args,**kwargs)


        [_,states] = self.ocp.sample(self.ocp.x,grid='control')
        [_,algebraics] = self.ocp.sample(self.ocp.z,grid='control')
        [_,controls] = self.ocp.sample(self.ocp.u,grid='control-')
        parameters_symbols = self.ocp.parameters['']+self.ocp.parameters['control']
        parameters = []
        for p in self.ocp.parameters['']:
            parameters.append(self.ocp.value(p))
        for p in self.ocp.parameters['control']:
            parameters.append(self.ocp.sample(p,grid='control-')[1])
        casadi_fun_name = 'ocpfun'
        is_coll = False
        if hasattr(self.ocp._method, "Xc"):
            is_coll = True

        variables = self.ocp.value(cs.vvcat(self.ocp.variables['']))
        [_,variables_control] = self.ocp.sample(cs.vvcat(self.ocp.variables['control']),grid='control-')
        [_,variables_states] = self.ocp.sample(cs.vvcat(self.ocp.variables['states']),grid='control')
        lam_g = self.ocp._method.opti.lam_g

        hotstart_symbol = cs.veccat(variables,variables_control,variables_states,lam_g)

        ocpfun = self.ocp.to_function(casadi_fun_name,
            [states]+(["z"] if is_coll else [cs.MX()])+[controls]+parameters+[hotstart_symbol],
            [states,algebraics,controls,hotstart_symbol],
            ['x0','z0','u0'] + [p.name() for p in self.ocp.parameters['']] + [p.name() for p in self.ocp.parameters['control']] + ['hotstart_in'],
            ['x','z','u','hotstart_out'])

        casadi_codegen_file_name_base = cs.name+"_codegen.c"
        casadi_codegen_file_name = os.path.join(cs.build_dir_abs,casadi_codegen_file_name_base)




    def casadi_call(self,funname,*args):
        method = f"casadi_c_{funname}_id"
        if self.use_codegen:
            if funname=="eval":
                method = f"{self.casadi_fun_name}"
            else:
                method = f"{self.casadi_fun_name}_{funname}"
        if not self.use_codegen:
            args = ("m->id",) + args
        return method + "(" + ",".join(args) + ")"

    def str_includes(self):
        for e in self.includes():
            return '#include "' + e + '"' + "\n"

    def includes(self):
        ret = []
        if self.use_codegen:
            ret.append(self.name + "_codegen.h")
        else:
            ret.append("casadi/casadi_c.h")
        return ret

    def setup_deps(self):
        if not self.use_codegen:
            yield MemoryField("id", "int")
            yield MemoryField("pop", "int"," Need to pop when destroyed?")
            yield LocalVariable("flag", "int")
        yield MemoryField("mem", "int")

    def str_setup(self):
        s = ""
        if not self.use_codegen:
            s+= f"""
            if (casadi_c_n_loaded()) {{ 
              m->id = casadi_c_id("{self.casadi_fun_name}");
              m->pop = 0;
            }}
            if (m->id<0) {{
              if (build_dir==0) {{
                strcpy(casadi_file, "{self.escape(self.build_dir_abs)}");
              }} else {{
                strcpy(casadi_file, build_dir);
              }}
              strcat(casadi_file, "/{self.name}.casadi");
              flag = casadi_c_push_file(casadi_file);
              m->pop = 1;
              if (flag) {{
                m->fatal(m, "initialize", "Could not load file '%s'.\\n", casadi_file);
                return 0;
              }}
              m->id = casadi_c_id("{self.casadi_fun_name}");
              if (m->id<0) {{
                m->fatal(m, "initialize", "Could not locate function with name '{self.casadi_fun_name}'.\\n");
                {self.call_destroy()}
                return 0;
              }}
            }}"""
        s += f"""/* Allocate memory (thread-safe) */
            {self.casadi_call("incref")};
            /* Checkout thread-local memory (not thread-safe) */
            m->mem = {self.casadi_call("checkout")};
            """
        return s
    def str_work(self):
        return self.casadi_call("work","sz_arg", "sz_res", "sz_iw", "sz_w")+";\n"

    def str_destroy(self):
        return f"""
               /* Free memory (thread-safe) */
              {self.casadi_call("decref")};
              /* Release thread-local (not thread-safe) */
              {self.casadi_call("release","m->mem")};
              {"" if self.use_codegen else "if (m->pop) casadi_c_pop();"}
              """

    def call_destroy(self):
        return f"{self.prefix}destroy(m);"

    def str_solve(self,args,rets):
        s= ""
        for i,e in enumerate(args):
            s+= f"m->arg_casadi[{i}] = {e};\n"
        for i,e in enumerate(rets):
            s+= f"m->res_casadi[{i}] = {e};\n"
        s += f"""
            return {self.casadi_call("eval","m->arg_casadi","m->res_casadi","m->iw_casadi","m->w_casadi","m->mem")};
        """
        return s

    def str_get_stats(self):
        return f"""return {self.casadi_call("stats") if self.use_codegen else 0};"""



###################################################################################################
#                                                                                                 #
# This file is part of HPIPM.                                                                     #
#                                                                                                 #
# HPIPM -- High-Performance Interior Point Method.                                                #
# Copyright (C) 2019 by Gianluca Frison.                                                          #
# Developed at IMTEK (University of Freiburg) under the supervision of Moritz Diehl.              #
# All rights reserved.                                                                            #
#                                                                                                 #
# The 2-Clause BSD License                                                                        #
#                                                                                                 #
# Redistribution and use in source and binary forms, with or without                              #
# modification, are permitted provided that the following conditions are met:                     #
#                                                                                                 #
# 1. Redistributions of source code must retain the above copyright notice, this                  #
#    list of conditions and the following disclaimer.                                             #
# 2. Redistributions in binary form must reproduce the above copyright notice,                    #
#    this list of conditions and the following disclaimer in the documentation                    #
#    and/or other materials provided with the distribution.                                       #
#                                                                                                 #
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND                 #
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED                   #
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE                          #
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR                 #
# ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES                  #
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;                    #
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND                     #
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT                      #
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS                   #
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.                                    #
#                                                                                                 #
# Author: Gianluca Frison, gianluca.frison (at) imtek.uni-freiburg.de                             #
#                                                                                                 #
###################################################################################################
from ctypes import *
import ctypes.util 
import numpy as np



class hpipm_ocp_qp_solver_arg2:


	def __init__(self, dim, mode):

		c_mode = 0
		if(mode=='speed_abs'):
			c_mode = 0
		elif(mode=='speed'):
			c_mode = 1
		elif(mode=='balance'):
			c_mode = 2
		elif(mode=='robust'):
			c_mode = 3
		else:
			raise NameError('hpipm_ocp_qp_solver_arg2: wrong mode')

		# save dim internally
		self.dim = dim

		# load hpipm shared library
		__hpipm   = CDLL('libhpipm.so')
		self.__hpipm = __hpipm

		# C qp struct
		arg_struct_size = __hpipm.d_ocp_qp_solver_arg_strsize()
		arg_struct = cast(create_string_buffer(arg_struct_size), c_void_p)
		self.arg_struct = arg_struct

		# C qp internal memory
		arg_mem_size = __hpipm.d_ocp_qp_solver_arg_memsize(dim.dim_struct)
		arg_mem = cast(create_string_buffer(arg_mem_size), c_void_p)
		self.arg_mem = arg_mem

		# create C qp
		__hpipm.d_ocp_qp_solver_arg_create(dim.dim_struct, arg_struct, arg_mem)

		# initialize default arguments
		__hpipm.d_ocp_qp_solver_arg_set_default(c_mode, dim.dim_struct, arg_struct)
	

	def set(self, field, value):
		if((field=='mu0') | (field=='alpha_min') | (field=='tol_stat') | (field=='tol_eq') | (field=='tol_ineq') | (field=='tol_comp') | (field=='reg_prim')):
			tmp_in = np.zeros((1,1))
			tmp_in[0][0] = value
			tmp = cast(tmp_in.ctypes.data, POINTER(c_double))
		elif((field=='iter_max') | (field=='pred_corr') | (field=='split_step') | (field=='reduce_eq_dof')):
			tmp_in = np.zeros((1,1), dtype=int)
			tmp_in[0][0] = value
			tmp = cast(tmp_in.ctypes.data, POINTER(c_int))
		else:
			raise NameError('hpipm_ocp_qp_solver_arg2.set: wrong field')
#		self.__hpipm.d_ocp_qp_ipm_arg_set_mu0.argtypes = [POINTER(c_double), c_void_p]
		field_name_b = field.encode('utf-8')
		self.__hpipm.d_ocp_qp_solver_arg_set(c_char_p(field_name_b), tmp, self.arg_struct)
		return


#	def codegen(self, file_name, mode):
#		file_name_b = file_name.encode('utf-8')
#		mode_b = mode.encode('utf-8')
#		self.__hpipm.d_ocp_qp_ipm_arg_codegen(c_char_p(file_name_b), c_char_p(mode_b), self.dim.dim_struct, self.arg_struct)
#		return


